# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.colors as mcolors
import datastock as ds

# specific
from . import _generic_check


# #############################################################################
# #############################################################################
#                     Main plotting
# #############################################################################


def plot_mesh(
    coll=None,
    key=None,
    ind_knot=None,
    ind_cent=None,
    crop=None,
    bck=None,
    nmax=None,
    color=None,
    dax=None,
    dmargin=None,
    fs=None,
    dleg=None,
    connect=None,
):
    """ Plot the desired mesh

    rect and tri meshes are constant
    polar meshes can vary in time

    """

    # --------------
    # check input

    (
     key, nd, mtype, ind_knot, ind_cent, crop, bck, color, dleg,
     return_neighbours,
     ) = _plot_mesh_check(
        coll=coll,
        key=key,
        ind_knot=ind_knot,
        ind_cent=ind_cent,
        crop=crop,
        bck=bck,
        color=color,
        dleg=dleg,
    )

    # ------------------------
    # call appropriate routine

    if nd == '1d':
        return plot_mesh_1d(
            coll=coll,
            key=key,
            ind_knot=ind_knot,
            ind_cent=ind_cent,
            #units=units,
            return_neighbours=return_neighbours,
            nmax=nmax,
            color=color,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dleg=dleg,
            connect=connect,
        )

    elif mtype in ['rect', 'tri']:
        # time-fixed meshes
        return _plot_mesh_2d_recttri(
            coll=coll,
            key=key,
            mtype=mtype,
            ind_knot=ind_knot,
            ind_cent=ind_cent,
            return_neighbours=return_neighbours,
            crop=crop,
            bck=bck,
            color=color,
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            dleg=dleg,
        )

    else:
        # possibly time-varying mesh
        raise NotImplementedError()
        # return _plot_mesh_2d_polar(
            # coll=coll,
            # key=key,
            # nmax=nmax,
            # color=color,
            # dax=dax,
            # fs=fs,
            # dmargin=dmargin,
            # dleg=dleg,
            # connect=connect,
        # )


# #############################################################################
# #############################################################################
#                           checks
# #############################################################################


def _plot_mesh_check(
    coll=None,
    key=None,
    ind_knot=None,
    ind_cent=None,
    crop=None,
    bck=None,
    color=None,
    dleg=None,
):

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=list(coll.dobj.get(coll._which_mesh, {}).keys()),
    )

    nd = coll.dobj[coll._which_mesh][key]['nd']
    mtype = coll.dobj[coll._which_mesh][key]['type']

    # crop, bck
    crop = ds._generic_check._check_var(crop, 'crop', default=True, types=bool)
    bck = ds._generic_check._check_var(bck, 'bck', default=True, types=bool)

    if mtype in ['rect', 'tri']:
        return_neighbours = True
    else:
        return_neighbours = False

    # ind_knot
    if ind_knot is not None:
        ind_knot = coll.select_mesh_elements(
            key=key, ind=ind_knot, elements='knots',
            returnas='data', return_neighbours=return_neighbours, crop=crop,
        )
        if return_neighbours is False:
            ind_knot = [ind_knot]

    # ind_cent
    if ind_cent is not None:
        ind_cent = coll.select_mesh_elements(
            key=key, ind=ind_cent, elements='cents',
            returnas='data', return_neighbours=return_neighbours, crop=crop,
        )
        if return_neighbours is False:
            ind_cent = [ind_cent]

    # color
    if color is None:
        color = 'k'
    if not mcolors.is_color_like(color):
        msg = (
            "Arg color must be a valid matplotlib color identifier!\n"
            f"Provided: {color}"
        )
        raise Exception(msg)

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    return (
        key, nd, mtype, ind_knot, ind_cent, crop, bck, color, dleg,
        return_neighbours,
    )


# #############################################################################
# #############################################################################
#                           prepare
# #############################################################################


def _plot_mesh_prepare_1d(
    coll=None,
    key=None,
    **kwd,
):

    # --------
    # prepare

    kknots = coll.dobj[coll._which_mesh][key]['knots'][0]
    knots = coll.ddata[kknots]['data']

    xx = np.array([knots, knots, np.full(knots.shape, np.nan)]).T.ravel()
    yy = np.array([
        np.zeros(knots.shape),
        np.ones(knots.shape),
        np.ones(knots.shape),
    ]).T.ravel()

    return xx, yy


def _plot_mesh_prepare_2d_rect(
    coll=None,
    key=None,
    crop=None,
    bck=None,
):

    # --------
    # prepare

    Rk, Zk = coll.dobj[coll._which_mesh][key]['knots']
    R = coll.ddata[Rk]['data']
    Z = coll.ddata[Zk]['data']

    vert = np.array([
        np.repeat(R, 3),
        np.tile((Z[0], Z[-1], np.nan), R.size),
    ])
    hor = np.array([
        np.tile((R[0], R[-1], np.nan), Z.size),
        np.repeat(Z, 3),
    ])

    # --------
    # compute

    grid_bck = None
    if crop is False or coll.dobj[coll._which_mesh][key]['crop'] is False:
        grid = np.concatenate((vert, hor), axis=1)

    else:

        crop = coll.ddata[coll.dobj[coll._which_mesh][key]['crop']]['data']

        grid = []
        icropR = np.r_[range(R.size-1), R.size-2]
        jcropZ = np.r_[range(Z.size-1), Z.size-2]

        # vertical lines  TBC
        for ii, ic in enumerate(icropR):
            if np.any(crop[ic, :]):
                if ii in [0, R.size-1]:
                    cropi = crop[ic, :]
                else:
                    cropi = crop[ic, :] | crop[ic-1, :]
                lseg = []
                for jj, jc in enumerate(jcropZ):
                    if jj == 0 and cropi[jc]:
                        lseg.append(Z[jj])
                    elif jj == Z.size-1 and cropi[jc]:
                        lseg.append(Z[jj])
                    elif cropi[jc] and not cropi[jc-1]:
                        if len(lseg) > 0:
                            lseg.append(np.nan)
                        lseg.append(Z[jj])
                    elif (not cropi[jc]) and cropi[jc-1]:
                        lseg.append(Z[jc])
                grid.append(np.concatenate(
                    (
                        np.array([R[ii]*np.ones((len(lseg),)), lseg]),
                        np.full((2, 1), np.nan)
                    ),
                    axis=1,
                ))

        # horizontal lines
        for jj, jc in enumerate(jcropZ):
            if np.any(crop[:, jc]):
                if jj in [0, Z.size-1]:
                    cropj = crop[:, jc]
                else:
                    cropj = crop[:, jc] | crop[:, jc-1]
                lseg = []
                for ii, ic in enumerate(icropR):
                    if ii in [0, R.size-1] and cropj[ic]:
                        lseg.append(R[ii])
                    elif cropj[ic] and not cropj[ic-1]:
                        if len(lseg) > 0:
                            lseg.append(np.nan)
                        lseg.append(R[ii])
                    elif (not cropj[ic]) and cropj[ic-1]:
                        lseg.append(R[ic])
                grid.append(np.concatenate(
                    (
                        np.array([lseg, Z[jj]*np.ones((len(lseg),))]),
                        np.full((2, 1), np.nan)
                    ),
                    axis=1,
                ))

        grid = np.concatenate(tuple(grid), axis=1)

        if bck is True:
            grid_bck = np.concatenate((vert, hor), axis=1)

    return grid, grid_bck


def _plot_mesh_prepare_2d_tri(
    coll=None,
    key=None,
    crop=None,
    bck=None,
):

    # --------
    # prepare

    grid_bck = None

    kknots = coll.dobj[coll._which_mesh][key]['knots']
    R = coll.ddata[kknots[0]]['data']
    Z = coll.ddata[kknots[1]]['data']

    indtri = coll.ddata[coll.dobj[coll._which_mesh][key]['ind']]['data']

    # find unique segments from all triangles
    segs = np.unique(
        np.sort(np.concatenate(
            (indtri[:, 0:2], indtri[:, 1:], indtri[:, ::2]),
            axis=0,
        )),
        axis=0,
    )

    # build long segments if possible
    ind = np.ones((segs.shape[0],), dtype=bool)
    ind[0] = False
    lseg = [segs[0, :]]
    last = segs[0, :]
    while np.any(ind):
        ii = segs[ind, 0] == last[-1]
        if np.any(ii):
            ii = ind.nonzero()[0][ii]
            dR0 = R[last[1]] - R[last[0]]
            dZ0 = Z[last[1]] - Z[last[0]]
            dR = np.diff(R[segs[ii, :]], axis=1)[:, 0]
            dZ = np.diff(Z[segs[ii, :]], axis=1)[:, 0]
            norm0 = np.sqrt(dR0**2 + dZ0**2)
            norm = np.sqrt(dR**2 + dZ**2)
            sca = (dR0*dR + dZ0*dZ) / (norm0 * norm)
            iwin = ii[np.argmax(sca)]
            lseg.append([segs[iwin, 1]])

        else:
            lseg.append([-1])
            iwin = ind.nonzero()[0][0]
            lseg.append(segs[iwin, :])

        last = segs[iwin, :]
        ind[iwin] = False

    lseg = np.concatenate(lseg)
    grid = np.array([R[lseg], Z[lseg]])
    grid[0, lseg == -1] = np.nan

    return grid, grid_bck


# def _plot_mesh_prepare_polar_cont(
    # coll=None,
    # key=None,
    # k2d=None,
    # RR=None,
    # ZZ=None,
    # ind=None,
    # nn=None,
# ):

    # # ---------------------
    # # sample mesh if needed

    # # ---------------------
    # # get map of rr / angle

    # if callable(k2d):

        # # check RR
        # if RR is None:
            # msg = (
                # "radius2d / angle2d are callable => provide RR and ZZ!"
            # )
            # raise Exception(msg)

        # # compute map
        # rr = k2d(RR, ZZ)[None, ...]
        # assert rr.ndim == RR.ndim + 1
        # reft = None
        # nt = 1

        # if nn is None:
            # nn = 50

        # # create vector
        # rad = np.linspace(np.nanmin(rr), np.nanmax(rr), nn)

    # else:
        # kn = coll.dobj[coll._which_mesh][key]['knots'][ind]
        # rad = coll.ddata[kn]['data']
        # kb2 = coll.ddata[k2d]['bsplines']

        # if RR is None:
            # km2 = coll.dobj['bsplines'][kb2]['mesh']
            # RR, ZZ = coll.get_sample_mesh(
                # key=km2,
                # res=None,
                # grid=True,
                # mode=None,
                # R=None,
                # Z=None,
                # DR=None,
                # DZ=None,
                # imshow=True,
            # )

        # rr = coll.interpolate_profile2d(
            # key=k2d,
            # R=RR,
            # Z=ZZ,
            # grid=False,
            # return_params=False,
        # )[0]

        # refr2d = coll.ddata[k2d]['ref']
        # refbs = coll.dobj['bsplines'][kb2]['ref']
        # if refr2d == refbs:
            # reft = None
            # nt = 1
            # rr = rr[None, ...]
        # elif len(refr2d) == len(refbs) + 1 and refr2d[1:] == refbs:
            # reft = refr2d[0]
            # nt = coll.dref[reft]['size']

    # assert rr.shape[0] == nt

    # # ----------------
    # # Compute contours

    # contR, contZ = _contours._get_contours(
        # xx0=RR,
        # xx1=ZZ,
        # val=rr,
        # levels=rad,
    # )

    # # refrad
    # refrad = coll.dobj[coll._which_mesh][key]['knots'][ind]

    # return contR, contZ, rad, reft, refrad, RR, ZZ


# def _plot_mesh_prepare_polar(
    # coll=None,
    # key=None,
    # # Necessary for callable radius2d
    # RR=None,
    # ZZ=None,
# ):

    # # --------
    # # prepare

    # # create rectangular grid and compute radius at each point
    # k2d = coll.dobj[coll._which_mesh][key]['radius2d']
    # (
        # contRrad, contZrad,
        # rad, reft, refrad,
        # RR, ZZ,
    # ) = _plot_mesh_prepare_polar_cont(
        # coll=coll,
        # key=key,
        # k2d=k2d,
        # RR=RR,
        # ZZ=ZZ,
        # ind=0,
        # nn=None,        # nrad if k2d callable
    # )

    # # -----------
    # # contour of angle if angle not None

    # contRang, contZang, ang, refang = None, None, None, None
    # if len(coll.dobj[coll._which_mesh][key]['shape-c']) == 2:
        # # create rectangular grid and compute radius at each point
        # k2d = coll.dobj[coll._which_mesh][key]['angle2d']
        # (
            # contRang, contZang,
            # ang, _, refang,
            # _, _,
        # ) = _plot_mesh_prepare_polar_cont(
            # coll=coll,
            # key=key,
            # k2d=k2d,
            # RR=RR,
            # ZZ=ZZ,
            # ind=1,
            # nn=None,        # nang if k2d callable
        # )

    # return (
        # contRrad, contZrad, rad, refrad,
        # contRang, contZang, ang, refang,
        # reft,
    # )


# #############################################################################
# #############################################################################
#                           mesh type specific
# #############################################################################


def plot_mesh_1d(
    coll=None,
    key=None,
    ind_knot=None,
    ind_cent=None,
    return_neighbours=None,
    units=None,
    nmax=None,
    color=None,
    dax=None,
    dmargin=None,
    fs=None,
    dleg=None,
    connect=None,
):
    """ Plot the desired spectral mesh

    """

    # --------------
    #  Prepare data

    xx, yy = _plot_mesh_prepare_1d(
        coll=coll,
        key=key,
    )

    # if units not in [None, 'eV']:
        # xx, _, _, cat = _spectralunits.convert_spectral(
            # data_in=xx,
            # units_in='eV',
            # units_out=units,
        # )
        # xlab = cat + r" ($" + units + "$)"

    # else:
        # xlab = r'energy ($eV$)'
    xlab = None


    # --------------
    # plot - prepare

    if dax is None:

        if dmargin is None:
            dmargin = {
                'left': 0.1, 'right': 0.9,
                'bottom': 0.1, 'top': 0.9,
                'hspace': 0.1, 'wspace': 0.1,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)
        ax0 = fig.add_subplot(gs[0, 0])
        ax0.set_xlabel(xlab)

        dax = {'spectral': ax0}

    dax = _generic_check._check_dax(dax=dax, main='spectral')

    # --------------
    # plot

    kax = 'spectral'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.plot(
            xx,
            yy,
            ls='-',
            lw=0.5,
            color=color,
            alpha=0.5,
            label=key,
        )

        if ind_knot is not None:
            ax.plot(
                ind_knot[0][0],
                0.5,
                marker='o',
                ms=8,
                ls='None',
                color=color,
                label='knots',
            )

            # if return_neighbours:
                # ax.plot(
                #     ind_knot[1][0, :, :],
                #     marker='x',
                #     ms=4,
                #     ls='None',
                #     color=color,
                # )

        if ind_cent is not None:
            ax.plot(
                ind_cent[0][0],
                0.5,
                marker='x',
                ms=8,
                ls='None',
                color=color,
                label='cents',
            )

            # if return_neighbours:
                # ax.plot(
                #     ind_cent[1][0, :, :],
                #     marker='o',
                #     ms=4,
                #     ls='None',
                #     color=color,
                # )

    # --------------
    # dleg

    if dleg is not False:
        for kax in dax.keys():
            dax[kax]['handle'].legend(**dleg)

    return dax


def _plot_mesh_2d_recttri(
    coll=None,
    key=None,
    mtype=None,
    ind_knot=None,
    ind_cent=None,
    return_neighbours=None,
    crop=None,
    bck=None,
    color=None,
    dax=None,
    fs=None,
    dmargin=None,
    dleg=None,
):

    # --------------
    #  Prepare data

    if mtype == 'rect':
        grid, grid_bck = _plot_mesh_prepare_2d_rect(
            coll=coll,
            key=key,
            crop=crop,
            bck=bck,
        )
    else:
        grid, grid_bck = _plot_mesh_prepare_2d_tri(
            coll=coll,
            key=key,
            crop=crop,
            bck=bck,
        )

    # --------------
    # plot - prepare

    if dax is None:

        if dmargin is None:
            dmargin = {
                'left': 0.1, 'right': 0.9,
                'bottom': 0.1, 'top': 0.9,
                'hspace': 0.1, 'wspace': 0.1,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)
        ax0 = fig.add_subplot(gs[0, 0], aspect='equal')
        ax0.set_xlabel(f'R (m)')
        ax0.set_ylabel(f'Z (m)')

        dax = {'cross': ax0}

    dax = _generic_check._check_dax(dax=dax, main='cross')

    # --------------
    # plot

    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if grid_bck is not None and bck is True:
            ax.plot(
                grid_bck[0, :],
                grid_bck[1, :],
                ls='-',
                lw=0.5,
                color=color,
                alpha=0.5,
                label=key,
            )

        ax.plot(
            grid[0, :],
            grid[1, :],
            color=color,
            ls='-',
            lw=1.,
            label=key,
        )

        if ind_knot is not None:

            ax.plot(
                ind_knot[0][0],
                ind_knot[0][1],
                marker='o',
                ms=8,
                ls='None',
                color=color,
                label='knots',
            )

            if return_neighbours:
                ax.plot(
                    ind_knot[1][0, :, :],
                    ind_knot[1][1, :, :],
                    marker='x',
                    ms=4,
                    ls='None',
                    color=color,
                )

        if ind_cent is not None:

            ax.plot(
                ind_cent[0][0],
                ind_cent[0][1],
                marker='x',
                ms=8,
                ls='None',
                color=color,
                label='cents',
            )

            if return_neighbours:
                ax.plot(
                    ind_cent[1][0, :, :],
                    ind_cent[1][1, :, :],
                    marker='o',
                    ms=4,
                    ls='None',
                    color=color,
                )

    # --------------
    # dleg

    if dleg is not False:
        for kax in dax.keys():
            dax[kax]['handle'].legend(**dleg)

    return dax


# def _plot_mesh_2d_polar(
    # coll=None,
    # key=None,
    # npts=None,
    # nmax=None,
    # color=None,
    # dax=None,
    # fs=None,
    # dmargin=None,
    # dleg=None,
    # connect=None,
# ):

    # # --------------
    # #  Prepare data

    # if nmax is None:
        # nmax = 2

    # (
        # contRrad, contZrad, rad, refrad,
        # contRang, contZang, ang, refang,
        # reft,
    # ) = _plot_mesh_prepare_polar(
        # coll=coll,
        # key=key,
    # )
    # refptsr = 'ptsr'
    # nt, nr, nptsr = contRrad.shape
    # if contRang is not None:
        # refptsa = 'ptsa'
        # _, nang, nptsa = contRang.shape

    # # --------------------
    # # Instanciate Plasma2D

    # coll2 = coll.__class__()

    # # ref
    # coll2.add_ref(
        # key=reft,
        # size=nt,
    # )
    # reft = list(coll2.dref.keys())[0]
    # coll2.add_ref(
        # key=refrad,
        # size=nr,
    # )
    # coll2.add_ref(
        # key=refptsr,
        # size=nptsr,
    # )

    # if contRang is not None:
        # coll2.add_ref(
            # key=refang,
            # size=nang,
        # )
        # coll2.add_ref(
            # key=refptsa,
            # size=nptsa,
        # )

    # # data
    # coll2.add_data(
        # key='radius',
        # data=rad,
        # ref=(refrad,)
    # )
    # coll2.add_data(
        # key='contRrad',
        # data=contRrad,
        # ref=(reft, refrad, refptsr)
    # )
    # coll2.add_data(
        # key='contZrad',
        # data=contZrad,
        # ref=(reft, refrad, refptsr)
    # )

    # if contRang is not None:
        # coll2.add_data(
            # key='angle',
            # data=ang,
            # ref=(refang,)
        # )
        # coll2.add_data(
            # key='contRang',
            # data=contRang,
            # ref=(reft, refang, refptsa)
        # )
        # coll2.add_data(
            # key='contZang',
            # data=contZang,
            # ref=(reft, refang, refptsa)
        # )

    # # -----
    # # plot

    # if contRang is None:
        # return coll2.plot_as_mobile_lines(
            # keyX='contRrad',
            # keyY='contZrad',
            # key_time=reft,
            # key_chan='radius',
            # connect=connect,
        # )

    # else:

        # daxrad, dgrouprad = coll2.plot_as_mobile_lines(
            # keyX='contRrad',
            # keyY='contZrad',
            # key_time=reft,
            # key_chan='radius',
            # connect=False,
            # inplace=False,
        # )

        # daxang, dgroupang = coll2.plot_as_mobile_lines(
            # keyX='contRang',
            # keyY='contZang',
            # key_time=reft,
            # key_chan='angle',
            # connect=False,
            # inplace=False,
        # )

        # # connect
        # if connect is False:
            # return (daxrad, daxang), (dgrouprad, dgroupang)

        # else:
            # daxrad.setup_interactivity(
                # kinter='inter0', dgroup=dgrouprad, dinc=None,
            # )
            # daxrad.disconnect_old()
            # daxrad.connect()

            # daxang.setup_interactivity(
                # kinter='inter0', dgroup=dgroupang, dinc=None,
            # )
            # daxang.disconnect_old()
            # daxang.connect()

            # daxrad.show_commands()
            # return daxrad, daxang



# -*- coding: utf-8 -*-


# Common
import numpy as np
import datastock as ds


from . import _generic_mesh
from . import _class01_checks_1d as _checks_1d


# #############################################################################
# #############################################################################
#                           mesh generic check
# #############################################################################


def check(
    coll=None,
    key=None,
    # knots
    knots0=None,
    knots1=None,
    # automated
    domain=None,
    res=None,
    # defined from pre-existing bsplines
    subkey0=None,
    subkey1=None,
    # attributes
    **kwdargs,
):

    # --------
    # keys

    # key
    key = ds._generic_check._obj_key(
        d0=coll._dobj.get(coll._which_mesh, {}),
        short='m',
        key=key,
    )

    # -------------
    # knots vectors

    knots0, knots1, res0, res1 = _mesh2DRect_check(
        key=key,
        knots0=knots0,
        knots1=knots1,
        # automated
        domain=domain,
        res=res,
    )
    cents0 = 0.5*(knots0[1:] + knots0[:-1])
    cents1 = 0.5*(knots1[1:] + knots1[:-1])

    # ------------------------
    # depend on other bsplines

    submesh0, subbs0, kwdargs = _checks_1d._defined_from(
        coll=coll,
        subkey=subkey0,
        # parameters
        kwdargs=kwdargs,
        nd='2d',
    )

    submesh1, subbs1, kwdargs = _checks_1d._defined_from(
        coll=coll,
        subkey=subkey1,
        # parameters
        kwdargs=kwdargs,
        nd='2d',
    )

    if subbs0 != subbs1:
        msg = (
            "Args subkey0 and subkey1 must be refering to "
            "2 data defined on the same 2d bsplines\n"
            "Provided:\n"
            "\t- subkey0: {subkey0}\n"
            "\t- subkey1: {subkey1}\n"
            "\t- subbs0: {subbs0}\n"
            "\t- subbs1: {subbs1}\n"
        )
        raise Exception(msg)

    subbs = subbs0
    submesh = submesh0

    # --------------
    # to dict

    dref, ddata, dobj = _to_dict(
        coll=coll,
        key=key,
        knots0=knots0,
        knots1=knots1,
        cents0=cents0,
        cents1=cents1,
        # sub quantity
        subkey0=subkey0,
        subkey1=subkey1,
        subbs=subbs,
        submesh=submesh,
        # attributes
        **kwdargs,
    )

    return key, dref, ddata, dobj


# #############################################################################
# #############################################################################
#                           check knots
# #############################################################################


def _mesh2DRect_check(
    key=None,
    knots0=None,
    knots1=None,
    domain=None,
    res=None,
):

    # --------------
    # check inputs

    # (domain, res) vs (knots0, knots1)
    lc = [
        knots0 is not None and knots1 is not None,
        domain is not None,
    ]
    if not any(lc):
        msg = (
            "Please at least (domain, res) or (knots0, knots1):\n"
            "Provided:\n"
            f"\t- domain, res: {domain}, {res}\n"
            f"\t- knots0, knots1: {knots0}, {knots1}\n"
        )
        raise Exception(msg)

    if lc[0]:

        # knots0
        knots0, res0 = _checks_1d._check_knots(
            key=key,
            knots=knots0,
            uniform=True,
        )

        # knots1
        knots1, res1 = _checks_1d._check_knots(
            key=key,
            knots=knots1,
            uniform=True,
        )

    elif lc[1]:
        # domain
        c0 = (
            isinstance(domain, list)
            and len(domain) == 2
            and all([
                hasattr(dd, '__iter__') and len(dd) >= 2 for dd in domain
            ])
        )
        if not c0:
            msg = (
                "Arg domain must be a list of 2 iterables of len() >= 2\n"
                f"Provided: {domain}"
            )
            raise Exception(msg)

        # res
        c0 = (
            res is None
            or np.isscalar(res)
            or isinstance(res, list) and len(res) == 2
        )
        if not c0:
            msg = (
                "Arg res must be a int, float or array or a list of 2 such\n"
                f"Provided: {res}"
            )
            raise Exception(msg)

        if np.isscalar(res) or res is None:
            res = [res, res]

        # -------------
        # derive knots

        knots0, res0, _ = _mesh2DRect_X_check(domain[0], res=res[0])
        knots1, res1, _ = _mesh2DRect_X_check(domain[1], res=res[1])

    return knots0, knots1, res0, res1


def _mesh2DRect_X_check(
    x=None,
    res=None,
):
    """ Returns knots (x) and associated resolution

    res can be:
        - int: numbr of mesh elements desired between knots
        - float: desired average mesh element size
        - array of floats: (one for each x, desired approximate mesh size)

    """

    # ------------
    # Check inputs

    # x
    try:
        x = np.unique(np.ravel(x).astype(float))
    except Exception as err:
        msg = f"x must be convertible to a sorted flat array of floats!\n{err}"
        raise Exception(msg)

    # res
    if res is None:
        res = 10

    lc = [
        isinstance(res, (int, np.int64, np.int32)) and len(x) == 2,
        isinstance(res, (float, np.floating)) and len(x) == 2,
        isinstance(res, (list, tuple, np.ndarray)) and len(x) == len(res),
    ]
    if not any(lc):
        msg = (
            "Arg res must be:\n"
            "\t- int: nb of mesh elements along x\n"
            "\t       requires len(x) = 2\n"
            "\t- float: approximate desired mesh element size along x\n"
            "\t       requires len(x) = 2\n"
            "\t- iterable: approximate desired mesh element sizes along x\n"
            "\t       requires len(x) = len(res)\n"
        )
        raise Exception(msg)

    if lc[0]:
        x_new = np.linspace(x[0], x[1], int(res)+1)
        res_new = res
        indsep = None

    elif lc[1]:
        nb = int(np.ceil((x[1]-x[0]) / res))
        x_new = np.linspace(x[0], x[1], nb+1)
        res_new = np.mean(np.diff(x))
        indsep = None

    else:

        # check conformity
        res = np.ravel(res).astype(float)
        delta = np.diff(x)
        res_sum = res[:-1] + res[1:]
        ind = res_sum > delta + 1.e-14
        if np.any(ind):
            msg = (
                "Desired resolution is not achievable for the following:\n"
                f"res_sum: {res_sum[ind]}\n"
                f"delta  : {delta[ind]}"
            )
            raise Exception(msg)

        # compute nn
        # nn = how many pairs can fit in the interval
        npairs = np.round(delta/res_sum).astype(int)
        res_sum_new = delta / npairs

        fract = res[:-1] / res_sum

        res_new = [None for ii in range(len(x)-1)]
        x_new = [None for ii in range(len(x)-1)]
        for ii in range(len(x)-1):
            res_new[ii] = (
                res_sum_new[ii]
                * np.linspace(fract[ii], 1.-fract[ii], 2*npairs[ii])
            )
            if ii == 0:
                res_add = np.concatenate(([0], np.cumsum(res_new[ii])))
            else:
                res_add = np.cumsum(res_new[ii])
            x_new[ii] = x[ii] + res_add

        indsep = np.cumsum(npairs[:-1]*2)
        res_new = np.concatenate(res_new)
        x_new = np.concatenate(x_new)

    return x_new, res_new, indsep


# #############################################################################
# #############################################################################
#                           to_dict
# #############################################################################


def _to_dict(
    coll=None,
    key=None,
    knots0=None,
    knots1=None,
    cents0=None,
    cents1=None,
    res0=None,
    res1=None,
    # submesh
    subkey0=None,
    subkey1=None,
    subbs=None,
    submesh=None,
    # attributes
    **kwdargs,
):

    # --------------------
    # check / format input

    kkr0, kcr0, kk0, kc0 = _generic_mesh.names_knots_cents(
        key=key,
        knots_name='0',
    )
    kkr1, kcr1, kk1, kc1 = _generic_mesh.names_knots_cents(
        key=key,
        knots_name='1',
    )

    # attributes
    latt = ['dim', 'quant', 'name', 'units']
    dim, quant, name, units = _generic_mesh._get_kwdargs_2d(kwdargs, latt)

    variable = not (np.isscalar(res0) and np.isscalar(res1))

    # subkey
    if subkey0 is not None:
        subkey = (subkey0, subkey1)
    else:
        subkey = None

    # --------------------
    # prepare dict

    # dref
    dref = {
        kkr0: {
            'size': knots0.size,
        },
        kkr1: {
            'size': knots1.size,
        },
        kcr0: {
            'size': cents0.size,
        },
        kcr1: {
            'size': cents1.size,
        },
    }

    # ddata
    ddata = {
        kk0: {
            'data': knots0,
            'units': units[0],
            # 'source': None,
            'dim': dim[0],
            'quant': quant[0],
            'name': name[0],
            'ref': kkr0,
        },
        kk1: {
            'data': knots1,
            'units': units[1],
            # 'source': None,
            'dim': dim[1],
            'quant': quant[1],
            'name': name[1],
            'ref': kkr1,
        },
        kc0: {
            'data': cents0,
            'units': units[0],
            # 'source': None,
            'dim': dim[0],
            'quant': quant[0],
            'name': name[0],
            'ref': kcr0,
        },
        kc1: {
            'data': cents1,
            'units': units[1],
            # 'source': None,
            'dim': dim[1],
            'quant': quant[1],
            'name': name[1],
            'ref': kcr1,
        },
    }

    # dobj
    dobj = {
        coll._which_mesh: {
            key: {
                'nd': '2d',
                'type': 'rect',
                'knots': (kk0, kk1),
                'cents': (kc0, kc1),
                # 'ref-k': (kRk, kZk),
                # 'ref-c': (kRc, kZc),
                'shape-c': (cents0.size, cents1.size),
                'shape-k': (knots0.size, knots1.size),
                'variable': variable,
                'subkey': subkey,
                'subbs': subbs,
                'submesh': submesh,
                'crop': False,
            },
        }
    }

    # additional attributes
    for k0, v0 in kwdargs.items():
        if k0 not in latt:
            dobj[coll._which_mesh][key][k0] = v0

    return dref, ddata, dobj


def _from_croppoly(crop_poly=None, domain=None):

    # ------------
    # check inputs

    c0 = hasattr(crop_poly, '__iter__') and len(crop_poly) == 2
    lc = [
        crop_poly is None,
        (
            c0
            and isinstance(crop_poly, tuple)
            and crop_poly[0].__class__.__name__ == 'Config'
            and (isinstance(crop_poly[1], str) or crop_poly[1] is None)
        )
        or crop_poly.__class__.__name__ == 'Config',
        c0
        and all([
            hasattr(cc, '__iter__') and len(cc) == len(crop_poly[0])
            for cc in crop_poly[1:]
        ])
        and np.asarray(crop_poly).ndim == 2
    ]

    if not any(lc):
        msg = (
            "Arg config must be a Config instance!"
        )
        raise Exception(msg)

    # -------------
    # Get polyand domain

    if lc[0]:
        # trivial case
        poly = None

    else:

        # -------------
        # Get poly from input

        if lc[1]:
            # (config, structure name)

            if crop_poly.__class__.__name__ == 'Config':
                config = crop_poly
                key_struct = None
            else:
                config, key_struct = crop_poly

            # key_struct if None
            if key_struct is None:
                lk, ls = zip(*[
                    (ss.Id.Name, ss.dgeom['Surf']) for ss in config.lStructIn
                ])
                key_struct = lk[np.argmin(ls)]

            # poly
            poly = config.dStruct['dObj']['Ves'][key_struct].Poly_closed

        else:

            # make sure poly is np.ndarraya and closed
            poly = np.asarray(crop_poly).astype(float)
            if not np.allclose(poly[:, 0], poly[:, -1]):
                poly = np.concatenate((poly, poly[:, 0:1]), axis=1)

        # -------------
        # Get domain from poly

        if domain is None:
            domain = [
                [poly[0, :].min(), poly[0, :].max()],
                [poly[1, :].min(), poly[1, :].max()],
            ]

    return domain, poly



# -*- coding: utf-8 -*-


# Common
import numpy as np
import datastock as ds


# #############################################################################
# #############################################################################
#                          bins generic check
# #############################################################################


def check(
    coll=None,
    key=None,
    edges=None,
    # custom names
    key_cents=None,
    key_ref=None,
    # additional attributes
    **kwdargs,
):

    # --------
    # keys

    # key
    key = ds._generic_check._obj_key(
        d0=coll._dobj.get(coll._which_bins, {}),
        short='b',
        key=key,
    )

    # ------------
    # edges

    edges = ds._generic_check._check_flat1darray(
        edges, 'edges',
        dtype=float,
        unique=True,
        can_be_None=False,
    )

    nb = edges.size - 1
    cents = 0.5*(edges[:-1] + edges[1:])

    # --------------------
    # safety check on keys

    # key_ref
    defk = f"{key}_nb"
    lout = [k0 for k0, v0 in coll.dref.items() if v0['size'] != nb]
    key_ref = ds._generic_check._check_var(
        key_ref, 'key_ref',
        types=str,
        default=defk,
        excluded=lout,
    )

    # key_cents
    defk = f"{key}_c"
    lout = [
        k0 for k0, v0 in coll.ddata.items()
        if not (
            v0['shape'] == (nb,)
            and key_ref in coll.dref.keys()
            and v0['ref'] == (key_ref,)
            and v0['monot'] == (True,)
        )
    ]
    key_cents = ds._generic_check._check_var(
        key_cents, 'key_cents',
        types=str,
        default=defk,
        excluded=lout,
    )

    # --------------
    # to dict

    dref, ddata, dobj = _to_dict(
        coll=coll,
        key=key,
        edges=edges,
        nb=nb,
        cents=cents,
        # custom names
        key_cents=key_cents,
        key_ref=key_ref,
        # attributes
        **kwdargs,
    )

    return key, dref, ddata, dobj


# ##############################################################
# ###############################################################
#                           to_dict
# ###############################################################


def _to_dict(
    coll=None,
    key=None,
    edges=None,
    nb=None,
    cents=None,
    # custom names
    key_cents=None,
    key_ref=None,
    # additional attributes
    **kwdargs,
):

    # attributes
    latt = ['dim', 'quant', 'name', 'units']
    dim, quant, name, units = [kwdargs.get(ss) for ss in latt]

    # -------------
    # prepare dict

    # dref
    if key_ref not in coll.dref.keys():
        dref = {
            key_ref: {
                'size': nb,
            },
        }
    else:
        dref = None

    # ddata
    if key_cents not in coll.ddata.keys():
        ddata = {
            key_cents: {
                'data': cents,
                'units': units,
                # 'source': None,
                'dim': dim,
                'quant': quant,
                'name': name,
                'ref': key_ref,
            },
        }
    else:
        ddata = None

    # dobj
    dobj = {
        coll._which_bins: {
            key: {
                'nd': '1d',
                'edges': edges,
                'cents': (key_cents,),
                'ref': (key_ref,),
                'shape': (nb,),
            },
        },
    }

    # additional attributes
    for k0, v0 in kwdargs.items():
        if k0 not in latt:
            dobj[coll._which_bins][key][k0] = v0

    return dref, ddata, dobj


# ##############################################################
# ###############################################################
#                   remove bins
# ###############################################################


def remove_bins(coll=None, key=None, propagate=None):

    # ----------
    # check

    # key
    wbins = coll._which_bins
    if wbins not in coll.dobj.keys():
        return

    if isinstance(key, str):
        key = [key]
    key = ds._generic_check._check_var_iter(
        key, 'key',
        types=(list, tuple),
        types_iter=str,
        allowed=coll.dobj.get(wbins, {}).keys(),
    )

    # propagate
    propagate = ds._generic_check._check_var(
        propagate, 'propagate',
        types=bool,
        default=True,
    )

    # ---------
    # remove

    for k0 in key:

        # specific data
        kdata = list(coll.dobj[wbins][k0]['cents'])
        coll.remove_data(kdata, propagate=propagate)

        # specific ref
        lref = list(coll.dobj[wbins][k0]['ref'])
        for rr in lref:
            if rr in coll.dref.keys():
                coll.remove_ref(rr, propagate=propagate)

        # obj
        coll.remove_obj(which=wbins, key=k0, propagate=propagate)



# -*- coding: utf-8 -*-


# Built-in
import warnings

# Common
import numpy as np
from scipy.spatial import ConvexHull
from matplotlib.path import Path
from contourpy import contour_generator
import datastock as ds


# tofu
from . import _generic_mesh
from . import _utils_bsplines
# from . import _class02_checks as _checks
from . import _class02_bsplines_rect
from . import _class02_bsplines_tri
from . import _class02_bsplines_polar
from . import _class02_bsplines_1d


# #############################################################################
# #############################################################################
#                           index
# #############################################################################


def _select_ind(
    coll=None,
    key=None,
    ind=None,
    elements=None,
    returnas=None,
    crop=None,
    # spatial vs spectral
):
    """ return indices
    ind can be:
        - None
        - tuple: (R, Z), possibly 2d
        - 'tuple-flat': (R, Z) flattened
        - np.ndarray: array of unique indices
        - 'array-flat': flattened ordered array of unique
    """

    # ------------
    # check inputs

    km, keybs, cat = _generic_mesh._get_key_mesh_vs_bplines(
        coll=coll,
        key=key,
    )

    meshtype = coll.dobj[coll._which_mesh][km]['type']

    shape2d = None
    if cat == coll._which_mesh:
        key = km
        shape2d = len(coll.dobj[cat][key]['shape-c']) == 2
    else:
        key = keybs
        shape2d = len(coll.dobj[cat][key]['shape']) == 2

    # ind, elements, ...
    # elements = cents or knots
    ind, elements, returnas, crop = _generic_mesh._select_ind_check(
        ind=ind,
        elements=elements,
        returnas=returnas,
        crop=crop,
        shape2d=shape2d,
    )

    elem = f'{elements}' if cat == coll._which_mesh else 'ref'

    if shape2d:
        if cat == coll._which_mesh:
            ke = f'shape-{elem[0]}'
            nR, nZ = coll.dobj[cat][key][ke]
        else:
            nR, nZ = coll.dobj[cat][key]['shape']
    else:
        if cat == coll._which_mesh:
            ke = f'shape-{elem[0]}'
            nelem = coll.dobj[cat][key][ke][0]
        else:
            nelem = coll.dobj[cat][key]['shape'][0]

    # ------------
    # ind to tuple

    if shape2d:
        ind_bool = np.zeros((nR, nZ), dtype=bool)
        if ind is None:
            # make sure R is varying in dimension 0
            ind_tup = (
                np.repeat(np.arange(0, nR)[:, None], nZ, axis=1),
                np.tile(np.arange(0, nZ), (nR, 1)),
            )
            ind_bool[...] = True

        elif isinstance(ind, tuple):
            c0 = (
                np.all((ind[0] >= 0) & (ind[0] < nR))
                and np.all((ind[1] >= 0) & (ind[1] < nZ))
            )
            if not c0:
                msg = (
                    f"Non-valid values in ind (< 0 or >= size ({nR}, {nZ}))"
                )
                raise Exception(msg)
            ind_tup = ind
            ind_bool[ind_tup[0], ind_tup[1]] = True

        else:
            if np.issubdtype(ind.dtype, np.integer):
                c0 = np.all((ind >= 0) & (ind < nR*nZ))
                if not c0:
                    msg = (
                        f"Non-valid values in ind (< 0 or >= size ({nR*nZ}))"
                    )
                    raise Exception(msg)
                ind_tup = (ind % nR, ind // nR)
                ind_bool[ind_tup[0], ind_tup[1]] = True

            elif np.issubdtype(ind.dtype, np.bool_):
                if ind.shape != (nR, nZ):
                    msg = (
                        f"Arg ind, if bool, must have shape {(nR, nZ)}\n"
                        f"Provided: {ind.shape}"
                    )
                    raise Exception(msg)
                # make sure R varies first
                ind_tup = ind.T.nonzero()[::-1]
                ind_bool = ind

            else:
                msg = f"Unknown ind dtype!\n\t- ind.dtype: {ind.dtype}"
                raise Exception(msg)

        if ind_tup[0].shape != ind_tup[1].shape:
            msg = (
                "ind_tup components do not have the same shape!\n"
                f"\t- ind_tup[0].shape = {ind_tup[0].shape}\n"
                f"\t- ind_tup[1].shape = {ind_tup[1].shape}"
            )
            raise Exception(msg)

    # triangular + polar1d case
    else:
        ind_bool = np.zeros((nelem,), dtype=bool)
        if ind is None:
            ind_bool[...] = True
        elif np.issubdtype(ind.dtype, np.integer):
            c0 = np.all((ind >= 0) & (ind < nelem))
            if not c0:
                msg = (
                    f"Arg ind has non-valid values (< 0 or >= size ({nelem}))"
                )
                raise Exception(msg)
            ind_bool[ind] = True
        elif np.issubdtype(ind.dtype, np.bool_):
            if ind.shape != (nelem,):
                msg = (
                    f"Arg ind, when array of bool, must have shape {(nelem,)}"
                    f"\nProvided: {ind.shape}"
                )
                raise Exception(msg)
            ind_bool = ind
        else:
            msg = (
                "Non-valid ind format!"
            )
            raise Exception(msg)

    # ------------
    # optional crop

    crop = (
        crop is True
        and coll.dobj[cat][key].get('crop') not in [None, False]
        and bool(np.any(~coll.ddata[coll.dobj[cat][key]['crop']]['data']))
    )
    if crop is True:
        cropi = coll.ddata[coll.dobj[cat][key]['crop']]['data']
        if meshtype == 'rect':
            if cat == coll._which_mesh and elements == 'knots':
                cropiknots = np.zeros(ind_bool.shape, dtype=bool)
                cropiknots[:-1, :-1] = cropi
                cropiknots[1:, :-1] = cropiknots[1:, :-1] | cropi
                cropiknots[1:, 1:] = cropiknots[1:, 1:] | cropi
                cropiknots[:-1, 1:] = cropiknots[:-1, 1:] | cropi

                ind_bool = ind_bool & cropiknots

                # ind_tup is not 2d anymore
                ind_tup = ind_bool.T.nonzero()[::-1]  # R varies first
                # warnings.warn("ind is not 2d anymore!")

            elif ind_tup[0].shape == cropi.shape:
                ind_bool = ind_bool & cropi
                # ind_tup is not 2d anymore
                ind_tup = ind_bool.T.nonzero()[::-1]  # R varies first
                # warnings.warn("ind is not 2d anymore!")

            else:
                ind_bool = ind_bool & cropi
                ind_tup = ind_bool.T.nonzero()[::-1]
        else:
            ind_bool &= cropi

    # ------------
    # tuple to return

    if returnas is bool:
        out = ind_bool
    elif returnas is int:
        out = ind_bool.nonzero()[0]
    elif returnas is tuple:
        out = ind_tup
    elif returnas == 'tuple-flat':
        # make sure R is varying first
        out = (ind_tup[0].T.ravel(), ind_tup[1].T.ravel())
    elif returnas is np.ndarray:
        out = ind_tup[0] + ind_tup[1]*nR
    elif returnas == 'array-flat':
        # make sure R is varying first
        out = (ind_tup[0] + ind_tup[1]*nR).T.ravel()
    else:
        out = ind_bool

    return out


# #############################################################################
# #############################################################################
#                           mesh elements
# #############################################################################


def _select_mesh(
    coll=None,
    key=None,
    ind=None,
    elements=None,
    returnas=None,
    return_ind_as=None,
    return_neighbours=None,
):
    """ ind is a tuple for rect """

    # ------------
    # check inputs

    wm = coll._which_mesh
    mtype = coll.dobj[wm][key]['type']

    (
        elements, returnas,
        return_ind_as, return_neighbours,
    ) = _generic_mesh._select_check(
        elements=elements,
        returnas=returnas,
        return_ind_as=return_ind_as,
        return_neighbours=return_neighbours,
    )

    # ------------
    # prepare

    if mtype in ['rect', 'tri']:
        kR, kZ = coll.dobj[wm][key][elements]
        R = coll.ddata[kR]['data']
        Z = coll.ddata[kZ]['data']
    else:
        kr = coll.dobj[wm][key][elements][0]
        rad = coll.ddata[kr]['data']

    # ------------
    # non-trivial case

    if returnas == 'ind':
        out = ind
    else:
        if mtype == 'rect':
            out = R[ind[0]], Z[ind[1]]
        elif mtype == 'tri':
            out = R[ind], Z[ind]
        else:
            out = rad[ind]

    # ------------
    # neighbours

    if return_neighbours is True:
        if mtype == 'rect':
            neigh = _select_mesh_neighbours_rect(
                coll=coll,
                key=key,
                ind=ind,
                elements=elements,
                returnas=returnas,
            )
        elif mtype == 'tri':
            neigh = _select_mesh_neighbours_tri(
                coll=coll,
                key=key,
                ind=ind,
                elements=elements,
                returnas=returnas,
                return_ind_as=return_ind_as,
            )
        else:
            neigh = _select_mesh_neighbours_1d(
                coll=coll,
                key=key,
                ind=ind,
                elements=elements,
                returnas=returnas,
                return_ind_as=return_ind_as,
            )

        return out, neigh
    else:
        return out


# TBF
def _select_mesh_neighbours_1d(
    coll=None,
    key=None,
    ind=None,
    elements=None,
    returnas=None,
    return_ind_as=None,
):
    """ ind is a bool

    if returnas = 'ind', ind is returned as a bool array
    (because the nb. of neighbours is not constant on a triangular mesh)

    """
    # ------------
    # neighbours

    elneig = 'cents' if elements == 'knots' else 'knots'
    kneig = coll.dobj[coll._which_mesh][key][f'{elneig}'][0]
    neig = coll.ddata[kneig]['data']
    nneig = neig.size

    # get tuple indices of neighbours
    ind = ind.nonzero()[0]
    shape = tuple(np.r_[ind.shape, 2].astype(int))
    ineig = np.zeros(shape, dtype=int)

    rsh = tuple(
        [2 if ii == len(shape)-1 else 1 for ii in range(len(shape))]
    )

    if elements == 'cents':
        ineig[...] = ind[..., None] + np.r_[0, 1].reshape(rsh)

    elif elements == 'knots':
        ineig[...] = ind[..., None] + np.r_[-1, 0].reshape(rsh)
        ineig[(ineig < 0) | (ineig >= nneig)] = -1

    # return neighbours in desired format
    if returnas == 'ind':
        neig_out = ineig
    else:
        neig_out = neig[ineig]
        neig_out[ineig == -1] = np.nan

    return neig_out



def _select_mesh_neighbours_rect(
    coll=None,
    key=None,
    ind=None,
    elements=None,
    returnas=None,
):
    """ ind is a tuple for rect """

    # ------------
    # neighbours

    elneig = 'cents' if elements == 'knots' else 'knots'
    kRneig, kZneig = coll.dobj[coll._which_mesh][key][f'{elneig}']
    Rneig = coll.ddata[kRneig]['data']
    Zneig = coll.ddata[kZneig]['data']
    nRneig = Rneig.size
    nZneig = Zneig.size

    # get tuple indices of neighbours
    shape = tuple(np.r_[ind[0].shape, 4].astype(int))
    neig = (
        np.zeros(shape, dtype=int),
        np.zeros(shape, dtype=int),
    )
    rsh = tuple(
        [4 if ii == len(shape)-1 else 1 for ii in range(len(shape))]
    )

    if elements == 'cents':
        neig[0][...] = ind[0][..., None] + np.r_[0, 1, 1, 0].reshape(rsh)
        neig[1][...] = ind[1][..., None] + np.r_[0, 0, 1, 1].reshape(rsh)
    elif elements == 'knots':
        neig[0][...] = ind[0][..., None] + np.r_[-1, 0, 0, -1].reshape(rsh)
        neig[1][...] = ind[1][..., None] + np.r_[-1, -1, 0, 0].reshape(rsh)
        neig[0][(neig[0] < 0) | (neig[0] >= nRneig)] = -1
        neig[1][(neig[1] < 0) | (neig[1] >= nZneig)] = -1

    # return neighbours in desired format
    if returnas == 'ind':
        neig_out = neig
    else:
        neig_out = np.array([Rneig[neig[0]], Zneig[neig[1]]])
        neig_out[:, (neig[0] == -1) | (neig[1] == -1)] = np.nan

    return neig_out


def _select_mesh_neighbours_tri(
    coll=None,
    key=None,
    ind=None,
    elements=None,
    returnas=None,
    return_ind_as=None,
):
    """ ind is a bool

    if returnas = 'ind', ind is returned as a bool array
    (because the nb. of neighbours is not constant on a triangular mesh)

    """
    # ------------
    # neighbours

    nind = ind.sum()
    kind = coll.dobj[coll._which_mesh][key]['ind']

    if returnas == 'data':
        elneig = 'cents' if elements == 'knots' else 'knots'
        kneig = coll.dobj[coll._which_mesh][key][elneig]
        Rneig = coll.ddata[kneig[0]]['data']
        Zneig = coll.ddata[kneig[1]]['data']

    if elements == 'cents':
        neig = coll.ddata[kind]['data'][ind, :]
        if returnas == 'ind':
            if return_ind_as is bool:
                kknots = coll.dobj[coll._which_mesh][key]['knots']
                nneig = coll.dref[f'{kknots}-ind']['size']
                neig_temp = np.zeros((nind, nneig), dtype=bool)
                for ii in range(nind):
                    neig_temp[ii, neig[ii, :]] = True
                neig = neig_temp
        else:
            neig = np.array([Rneig[neig], Zneig[neig]])
    else:
        ind_int = ind.nonzero()[0]
        neig = np.array([
            np.any(coll.ddata[kind]['data'] == ii, axis=1)
            for ii in ind_int
        ])
        c0 = returnas == 'ind' and return_ind_as is int
        if c0 or returnas == 'data':
            nmax = np.sum(neig, axis=1)
            if returnas == 'ind':
                neig_temp = -np.ones((nind, nmax.max()), dtype=int)
                for ii in range(nind):
                    neig_temp[ii, :nmax[ii]] = neig[ii, :].nonzero()[0]
            else:
                neig_temp = np.full((2, nind, nmax.max()), np.nan)
                for ii in range(nind):
                    neig_temp[0, ii, :nmax[ii]] = Rneig[neig[ii, :]]
                    neig_temp[1, ii, :nmax[ii]] = Zneig[neig[ii, :]]
            neig = neig_temp

    return neig


# TBF
# def _select_mesh_neighbours_polar(
    # coll=None,
    # key=None,
    # ind=None,
    # elements=None,
    # returnas=None,
    # return_ind_as=None,
# ):
    # """ ind is a bool

    # if returnas = 'ind', ind is returned as a bool array
    # (because the nb. of neighbours is not constant on a triangular mesh)

    # """

    # elneig = 'cents' if elements == 'knots' else 'knots'
    # kneig = coll.dobj[coll._which_mesh][key][f'{elneig}']
    # rneig = coll.ddata[kneig[0]]['data']
    # nrneig = rneig.size


    # # ----------------
    # # radius + angle

    # if len(kneig) == 2:
        # aneig = coll.ddata[kneig[1]]['data']
        # naneig = aneig.size

        # # prepare indices
        # shape = tuple(np.r_[ind[0].shape, 2])
        # neig = (
            # np.zeros((nrneig, 2), dtype=bool),
            # np.zeros((naneig, 2), dtype=bool),
        # )

        # # get indices of neighbours
        # if elements == 'cents':
            # neig[0][...] = ind[0][..., None] + np.r_[0, 1, 1, 0].reshape(rsh)
            # neig[1][...] = ind[1][..., None] + np.r_[0, 0, 1, 1].reshape(rsh)
        # elif elements == 'knots':
            # neig[0][...] = ind[0][..., None] + np.r_[-1, 0, 0, -1].reshape(rsh)
            # neig[1][...] = ind[1][..., None] + np.r_[-1, -1, 0, 0].reshape(rsh)
            # neig[0][(neig[0] < 0) | (neig[0] >= nRneig)] = -1
            # neig[1][(neig[1] < 0) | (neig[1] >= nZneig)] = -1


    # # ----------------
    # # radius only

    # else:
        # # prepare indices
        # neig = np.zeros((nrneig, 2), dtype=bool)

        # # get indices of neighbours
        # if elements == 'cents':
            # neig[0][...] = ind[0][..., None] + np.r_[0, 1, 1, 0].reshape(rsh)
            # neig[1][...] = ind[1][..., None] + np.r_[0, 0, 1, 1].reshape(rsh)

    # # return neighbours in desired format
    # if returnas == 'ind':
        # neig_out = neig
    # else:
        # if len(kneig) == 2:
            # neig_out = np.array([rneig[neig[0]], zneig[neig[1]]])
            # neig_out[:, (neig[0] == -1) | (neig[1] == -1)] = np.nan
        # else:
            # neig_out = rneig[neig]

    # return neig_out


# #############################################################################
# #############################################################################
#                           bsplines
# #############################################################################


def _select_bsplines(
    coll=None,
    key=None,
    ind=None,
    returnas=None,
    return_cents=None,
    return_knots=None,
    crop=None,
):
    """ ind is a tuple """

    # ------------
    # check inputs

    wm = coll._which_mesh
    wbs = coll._which_bsplines

    # check key
    keym, keybs, _ = _generic_mesh._get_key_mesh_vs_bplines(
        coll=coll,
        key=key,
        which=wbs,
    )

    nd = coll.dobj[wm][keym]['nd']
    mtype = coll.dobj[wm][keym]['type']

    # ----
    # ind

    if mtype == 'rect':
        returnasind = tuple
    else:
        returnasind = bool

    ind = _select_ind(
        coll=coll,
        key=key,
        ind=ind,
        elements=None,
        returnas=returnasind,
        crop=crop,
    )

    # ------------
    # knots, cents

    if mtype == 'rect':
        kRk, kZk = coll.dobj[wm][keym]['knots']
        kRc, kZc = coll.dobj[wm][keym]['cents']

        out = _mesh2DRect_bsplines_knotscents(
            returnas=returnas,
            return_knots=return_knots,
            return_cents=return_cents,
            ind=ind,
            deg=coll.dobj[wbs][key]['deg'],
            Rknots=coll.ddata[kRk]['data'],
            Zknots=coll.ddata[kZk]['data'],
            Rcents=coll.ddata[kRc]['data'],
            Zcents=coll.ddata[kZc]['data'],
        )

    elif mtype == 'tri':
        clas = coll.dobj[wbs][key]['class']
        out = clas._get_knotscents_per_bs(
            returnas=returnas,
            return_knots=return_knots,
            return_cents=return_cents,
            ind=ind,
        )

    else:
        kk = coll.dobj[wm][keym]['knots'][0]
        kc = coll.dobj[wm][keym]['cents'][0]

        out = _mesh1d_bsplines_knotscents(
            returnas=returnas,
            return_knots=return_knots,
            return_cents=return_cents,
            ind=ind,
            deg=coll.dobj[wbs][key]['deg'],
            knots=coll.ddata[kk]['data'],
            cents=coll.ddata[kc]['data'],
        )

    # ------------
    # return

    if return_cents is True and return_knots is True:
        return ind, out[0], out[1]
    elif return_cents is True or return_knots is True:
        return ind, out
    else:
        return ind


def _mesh1d_bsplines_knotscents(
    returnas=None,
    return_knots=None,
    return_cents=None,
    ind=None,
    deg=None,
    knots=None,
    cents=None,
):

    # -------------
    # check inputs

    return_knots = ds._generic_check._check_var(
        return_knots, 'return_knots',
        types=bool,
        default=True,
    )
    return_cents = ds._generic_check._check_var(
        return_cents, 'return_cents',
        types=bool,
        default=True,
    )
    if return_knots is False and return_cents is False:
        return

    # -------------
    # compute

    if return_knots is True:

        knots_per_bs = _utils_bsplines._get_knots_per_bs(
            knots, deg=deg, returnas=returnas,
        )
        if ind is not None:
            knots_per_bs = knots_per_bs[:, ind]

    if return_cents is True:

        cents_per_bs = _utils_bsplines._get_cents_per_bs(
            cents, deg=deg, returnas=returnas,
        )
        if ind is not None:
            cents_per_bs = cents_per_bs[:, ind]

    # -------------
    # return

    if return_knots is True and return_cents is True:
        out = (knots_per_bs, cents_per_bs)
    elif return_knots is True:
        out = knots_per_bs
    else:
        out = cents_per_bs
    return out


def _mesh2DRect_bsplines_knotscents(
    returnas=None,
    return_knots=None,
    return_cents=None,
    ind=None,
    deg=None,
    Rknots=None,
    Zknots=None,
    Rcents=None,
    Zcents=None,
):

    # -------------
    # check inputs

    return_knots = ds._generic_check._check_var(
        return_knots, 'return_knots',
        types=bool,
        default=True,
    )
    return_cents = ds._generic_check._check_var(
        return_cents, 'return_cents',
        types=bool,
        default=True,
    )
    if return_knots is False and return_cents is False:
        return

    # -------------
    # compute

    if return_knots is True:

        knots_per_bs_R = _utils_bsplines._get_knots_per_bs(
            Rknots, deg=deg, returnas=returnas,
        )
        knots_per_bs_Z = _utils_bsplines._get_knots_per_bs(
            Zknots, deg=deg, returnas=returnas,
        )
        if ind is not None:
            knots_per_bs_R = knots_per_bs_R[:, ind[0]]
            knots_per_bs_Z = knots_per_bs_Z[:, ind[1]]

        nknots = knots_per_bs_R.shape[0]
        knots_per_bs_R = np.tile(knots_per_bs_R, (nknots, 1))
        knots_per_bs_Z = np.repeat(knots_per_bs_Z, nknots, axis=0)

    if return_cents is True:

        cents_per_bs_R = _utils_bsplines._get_cents_per_bs(
            Rcents, deg=deg, returnas=returnas,
        )
        cents_per_bs_Z = _utils_bsplines._get_cents_per_bs(
            Zcents, deg=deg, returnas=returnas,
        )
        if ind is not None:
            cents_per_bs_R = cents_per_bs_R[:, ind[0]]
            cents_per_bs_Z = cents_per_bs_Z[:, ind[1]]

        ncents = cents_per_bs_R.shape[0]
        cents_per_bs_R = np.tile(cents_per_bs_R, (ncents, 1))
        cents_per_bs_Z = np.repeat(cents_per_bs_Z, ncents, axis=0)

    # -------------
    # return

    if return_knots is True and return_cents is True:
        out = (
            (knots_per_bs_R, knots_per_bs_Z), (cents_per_bs_R, cents_per_bs_Z)
        )
    elif return_knots is True:
        out = (knots_per_bs_R, knots_per_bs_Z)
    else:
        out = (cents_per_bs_R, cents_per_bs_Z)
    return out



# -*- coding: utf-8 -*-


# Common
import numpy as np
import datastock as ds


from . import _generic_mesh


_ELEMENTS = 'knots'


# ################################################################
# ###############################################################
#                           mesh generic check
# ###############################################################

"""
def _mesh2D_polar_check(
    coll=None,
    radius=None,
    angle=None,
    radius2d=None,
    angle2d=None,
    key=None,
    # parameters
    radius_dim=None,
    radius_quant=None,
    radius_name=None,
    radius_units=None,
    angle_dim=None,
    angle_quant=None,
    angle_name=None,
):

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        excluded=list(coll.dobj.get('mesh', {}).keys())
    )

    # --------------------
    # check / format input

    krk, krc, kkr, kcr = _generic_mesh.names_knots_cents(key=key, knots_name='r')
    kak, kac, kka, kca = _generic_mesh.names_knots_cents(key=key, knots_name='ang')

    # radius data
    radius, _, _ = _mesh1D_check(
        x=radius,
        x_name='radius',
        uniform=False,
    )

    # angle data
    c0 = (
        angle is None
        or (
            hasattr(angle, '__iter__')
            and np.asarray(angle).ndim == 1
            and np.unique(angle).size == np.array(angle).size
            and np.allclose(
                np.unique(np.arctan2(np.sin(angle), np.cos(angle))),
                angle,
            )
        )
    )
    if not c0:
        msg = (
            "Arg angle either\n:"
            "\t- None: radial-only polar mesh"
            "\t- convertible to a 1d increasing array\n"
            "\t\t it must be in radians\n"
            "\t\t it must be in the [-pi; pi] interval\n"
            f"\t- Provided: {angle}"
        )
        raise Exception(msg)

    # extract data
    rknot = np.unique(radius)
    rcent = 0.5*(rknot[1:] + rknot[:-1])

    # radius2d
    dradius = _check_polar_2dquant(
        coll=coll,
        quant2d=radius2d,
        quant2d_name='radius2d',
        dim=radius_dim,
        quant=radius_quant,
        name=radius_name,
        units=radius_units,
    )

    if callable(radius2d):
        keysm = None
    else:
        keysm = coll.dobj['bsplines'][coll.ddata[radius2d]['bsplines']]['mesh']

    if angle is not None:
        aknot = np.unique(np.arctan2(np.sin(angle), np.cos(angle)))
        acent = 0.5*(aknot[1:] + aknot[:-1])
        amid = 0.5*(aknot[-1] + (2.*np.pi + aknot[0]))
        amid = np.arctan2(np.sin(amid), np.cos(amid))
        if amid < acent[0]:
            acent = np.r_[amid, acent]
        else:
            acent = np.r_[acent, amid]

    # -------
    # angle2d

    if angle2d is not None:
        dangle = _check_polar_2dquant(
            coll=coll,
            quant2d=angle2d,
            quant2d_name='angle2d',
            dim=angle_dim,
            quant=angle_quant,
            name=angle_name,
            units='rad',
        )

        # check angle units = rad
        if dangle['units'] != 'rad':
            msg = (
                "Angle units must be rad\n"
                f"\t Provided: {dangle['units']}"
            )
            raise Exception(msg)

        # check angle2d is like radius2d
        c0 = (
            (callable(radius2d) and callable(angle2d))
            or coll._ddata[radius2d]['ref'] == coll._ddata[angle2d]['ref']
        )
        if not c0:
            msg = (
                "radius2d and angle2d must be of the same type, either:\n"
                "\t- both callable\n"
                "\t- both data keys with identical ref!\n"
                f"Provided:\n"
                f"\t- radius2d: {radius2d}\n"
                f"\t- angle2d: {angle2d}\n"
            )
            raise Exception(msg)

    # --------------------
    # prepare dict

    # dref
    dref = {
        krk: {'size': rknot.size},
        krc: {'size': rcent.size},
    }

    if angle is not None:
        dref.update({
            kak: {
                'size': aknot.size,
            },
            kac: {
                'size': acent.size,
            },
        })

    # ddata
    ddata = {
        kkr: {
            'data': rknot,
            'ref': krk,
            **dradius,
        },
        kcr: {
            'data': rcent,
            'ref': krc,
            **dradius,
        },
    }

    if angle is not None:
        ddata.update({
            kka: {
                'data': aknot,
                'ref': kak,
                **dangle,
            },
            kca: {
                'data': acent,
                'ref': kac,
                **dangle,
            },
        })

    # dobj
    if angle is None:
        dmesh = {
            key: {
                'type': 'polar',
                'knots': (kkr,),
                'cents': (kcr,),
                'shape-c': rcent.shape,
                'shape-k': rknot.shape,
                'radius2d': radius2d,
                'angle2d': angle2d,
                'submesh': keysm,
                'crop': False,
            },
        }
    else:
        dmesh = {
            key: {
                'type': 'polar',
                'knots': (kkr, kka),
                'cents': (kcr, kca),
                'shape-c': (rcent.size, acent.size),
                'shape-k': (rknot.size, aknot.size),
                'radius2d': radius2d,
                'angle2d': angle2d,
                'submesh': keysm,
                'crop': False,
            },
        }

    return dref, ddata, dmesh


def _check_polar_2dquant(
    quant2d=None,
    coll=None,
    quant2d_name=None,
    # parameters
    dim=None,
    quant=None,
    name=None,
    units=None,
):

    if coll.dobj.get('bsplines') is not None:
        lok = [
            k0 for k0, v0 in coll.ddata.items()
            if v0.get('bsplines') in coll.dobj['bsplines'].keys()
        ]
    else:
        lok = []

    lc = [
        callable(quant2d),
        isinstance(quant2d, str) and quant2d in lok
    ]
    if not any(lc):
        msg = (
            f"Arg {quant2d_name} must be either:\n"
            f"\t- callable: {quant2d_name} = func(R, Z)\n"
            f"\t- key to existing 2d data in {lok}\n"
            f"Provided: {quant2d}\n"
        )
        raise Exception(msg)

    # quantities
    dquant = {'dim': dim, 'quant': quant, 'name': name, 'units': units}
    if isinstance(quant2d, str):
        for k0 in dquant.keys():
            if dquant[k0] is None:
                dquant[k0] = str(coll.ddata[quant2d][k0])

    return dquant
"""



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import scipy.sparse as scpsp
import datastock as ds


from . import _utils_bsplines_operators as _operators
from . import _class02_bsplines_operators_1d as _operators_1d


_LOPERATORS_INT = [
    'D1',
    'D2',
    'D3',
    'D0N1',
    'D0N2',
    'D1N2',
    'D2N2',
    'D3N2',
]


# ###############################################################
# ###############################################################
#                   Mesh2DRect - bsplines - operators
# ###############################################################


def get_mesh2dRect_operators(
    operator=None,
    geometry=None,
    deg=None,
    knotsx_mult=None,
    knotsy_mult=None,
    knotsx_per_bs=None,
    knotsy_per_bs=None,
    overlap=None,
    sparse_fmt=None,
    cropbs_flat=None,
    # specific to deg = 0
    cropbs=None,
    centered=None,
    # for D1N2 only (to feed to tomotok / Mfr)
    returnas_element=None,
):

    # ------------
    # check inputs

    (
        operator, geometry, sparse_fmt, returnas_element,
    ) = _operators_1d._check(
        deg=deg,
        operator=operator,
        geometry=geometry,
        sparse_fmt=sparse_fmt,
        returnas_element=returnas_element,
    )

    # ------------
    # prepare

    nx, ny = knotsx_per_bs.shape[1], knotsy_per_bs.shape[1]
    kR = np.tile(knotsx_per_bs, ny)
    kZ = np.repeat(knotsy_per_bs, nx, axis=1)
    nbs = nx*ny

    if cropbs_flat is None:
        cropbs_flat = False
    if cropbs_flat is not False:
        c0 = (
            isinstance(cropbs_flat, np.ndarray)
            and cropbs_flat.shape == (nbs,)
            and cropbs_flat.dtype == np.bool_
        )
        if not c0:
            msg = (
                f"Arg cropbs_flat must be a bool array of shape {(nbs,)}\n"
                f"Provided: {cropbs_flat.shape}"
            )
            raise Exception(msg)
        nbscrop = cropbs_flat.sum()
        shape = (nbscrop, nbscrop)
        indbs = -np.ones((nbs,), dtype=int)
        indbs[cropbs_flat] = np.arange(0, nbscrop)
    else:
        shape = (nbs, nbs)
        indbs = np.arange(0, nbs)

    if 'N2' in operator and deg >= 1:
        # get intersection indices array
        if cropbs_flat is False:
            nbtot = np.sum(overlap >= 0)
        else:
            ind = cropbs_flat[None, :] & cropbs_flat[overlap]
            nbtot = np.sum(ind)

        # prepare data and indices arrays
        if operator == 'D0N2':
            data = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)
        elif operator == 'D1N2':
            datadR = np.full((nbtot,), np.nan)
            datadZ = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)
        elif operator == 'D2N2':
            datad2R = np.full((nbtot,), np.nan)
            datad2Z = np.full((nbtot,), np.nan)
            datadRZ = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)

    # ------------
    # D0 - integral

    geom = geometry
    if operator == 'D0N1':

        if deg == 0:

            opZ = _operators._D0N1_Deg0(kZ)
            opR = _operators._D0N1_Deg0(kR, geom)
            opmat = opR * opZ

        elif deg == 1:

            opZ = _operators._D0N1_Deg1(kZ)
            opR = _operators._D0N1_Deg1(kR, geom)
            opmat = opR * opZ

        elif deg == 2:

            iZ1, iZ21, iZ22, iZ3 = _operators._D0N1_Deg2(kZ)
            iR1, iR21, iR22, iR3 = _operators._D0N1_Deg2(kR, geom)
            opmat = (
                (iR1 + iR21 + iR22 + iR3)
                * (iZ1 + iZ21 + iZ22 + iZ3)
            )

        elif deg == 3:

            msg = "Integral D0N1 not implemented for deg=3 yet!"
            raise NotImplementedError(msg)

        # crop
        if cropbs_flat is not False:
            opmat = opmat[cropbs_flat]

    # ------------
    # D1 - gradient

    elif operator == 'D1':

        # Treat separately discrete case
        if deg == 0:
            gradR, gradZ = _operators._D1_Deg0_2d(
                knotsx_mult=knotsx_mult,
                knotsy_mult=knotsy_mult,
                cropbs=cropbs,
                cropbs_flat=cropbs_flat,
                nx=nx,
                ny=ny,
                nbs=nbs,
                centered=centered,
            )
            opmat = (
                scpsp.csc_matrix(gradR),
                scpsp.csc_matrix(gradZ),
            )

        elif deg >= 1:
            raise NotImplementedError()

    # ------------
    # D0N2

    elif operator == 'D0N2' and deg == 0:

        iZ = _operators._D0N2_Deg0(kZ, geom)
        iR = _operators._D0N2_Deg0(kR, geom)

        if cropbs_flat is not False:
            iR = iR[cropbs_flat]
            iZ = iZ[cropbs_flat]

        opmat = scpsp.diags(
            [iR*iZ],
            [0],
            shape=None,
            format=sparse_fmt,
            dtype=float,
        )

    elif operator == 'D0N2':

        # pre-compute integrals
        if deg == 1:
            iR = _operators._D0N2_Deg1(knotsx_mult, geom)
            iZ = _operators._D0N2_Deg1(knotsy_mult)

        elif deg == 2:
            iR = _operators._D0N2_Deg2(knotsx_mult, geom)
            iZ = _operators._D0N2_Deg2(knotsy_mult)

        elif deg == 3:
            msg = "Integral D0N2 not implemented for deg=3!"
            raise NotImplementedError(msg)

        # set non-diagonal elements
        i0 = 0
        for ir in range(nx):
            for iz in range(ny):

                iflat = ir + iz*nx
                if cropbs_flat is not False and not cropbs_flat[iflat]:
                    continue

                # general case
                overlapi = overlap[:, iflat][overlap[:, iflat] > iflat]

                # diagonal element
                data[i0] = iR[0, ir] * iZ[0, iz]
                row[i0] = indbs[iflat]
                column[i0] = indbs[iflat]
                i0 += 1

                # non-diagonal elements (symmetric)
                for jflat in overlapi:

                    if cropbs_flat is not False and not cropbs_flat[jflat]:
                        continue

                    jr = jflat % nx
                    jz = jflat // nx

                    # store (i, j) and (j, i) (symmetric matrix)
                    if jr >= ir:
                        iiR = iR[jr - ir, ir]
                    else:
                        iiR = iR[abs(jr - ir), jr]
                    if jz >= iz:
                        iiZ = iZ[jz - iz, iz]
                    else:
                        iiZ = iZ[abs(jz - iz), jz]

                    data[i0:i0+2] = iiR * iiZ
                    row[i0:i0+2] = (indbs[iflat], indbs[jflat])
                    column[i0:i0+2] = (indbs[jflat], indbs[iflat])
                    i0 += 2

        assert i0 == nbtot
        opmat = scpsp.csc_matrix((data, (row, column)), shape=shape)

    # ------------
    # D1N2

    elif operator == 'D1N2':

        # Treat separately discrete case
        if deg == 0:
            gradR, gradZ = _operators._D1_Deg0_2d(
                knotsx_mult=knotsx_mult,
                knotsy_mult=knotsy_mult,
                cropbs=cropbs,
                cropbs_flat=cropbs_flat,
                nx=nx,
                ny=ny,
                nbs=nbs,
            )

            # surface elements
            dZ = np.repeat(knotsy_mult[1:] - knotsy_mult[:-1], nx)
            if geometry == 'linear':
                dR = np.tile(
                    knotsx_mult[1:] - knotsx_mult[:-1],
                    ny,
                )
            else:
                dR = np.tile(
                    0.5*(knotsx_mult[1:]**2 - knotsx_mult[:-1]**2),
                    ny,
                )

            dS = dR*dZ
            if cropbs_flat is not False:
                dS = dS[cropbs_flat]

            # Does not seem to give positive definite matrix, to be checked ?
            if returnas_element is True:
                opmat = (
                    scpsp.csc_matrix(gradR*np.sqrt(dS[:, None])),
                    scpsp.csc_matrix(gradZ*np.sqrt(dS[:, None])),
                )

            else:
                opmat = (
                    scpsp.csc_matrix((gradR.T.dot(gradR))*(dS[:, None])),
                    scpsp.csc_matrix((gradZ.T.dot(gradZ))*(dS[:, None])),
                )

        else:

            # pre-compute integrals for exact operator deg >= 1
            if deg == 1:
                idR = _operators._D1N2_Deg1(knotsx_mult, geom)
                idZ = _operators._D1N2_Deg1(knotsy_mult)
                iR = _operators._D0N2_Deg1(knotsx_mult, geom)
                iZ = _operators._D0N2_Deg1(knotsy_mult)
            elif deg == 2:
                idR = _operators._D1N2_Deg2(knotsx_mult, geom)
                idZ = _operators._D1N2_Deg2(knotsy_mult)
                iR = _operators._D0N2_Deg2(knotsx_mult, geom)
                iZ = _operators._D0N2_Deg2(knotsy_mult)
            elif deg == 3:
                msg = "Integral D1N2 not implemented for deg=3!"
                raise NotImplementedError(msg)

            # set non-diagonal elements
            i0 = 0
            for ir in range(nx):
                for iz in range(ny):

                    iflat = ir + iz*nx
                    if cropbs_flat is not False and not cropbs_flat[iflat]:
                        continue

                    # general case
                    overlapi = overlap[:, iflat][overlap[:, iflat] > iflat]

                    # diagonal element
                    datadR[i0] = idR[0, ir] * iZ[0, iz]
                    datadZ[i0] = iR[0, ir] * idZ[0, iz]
                    row[i0] = indbs[iflat]
                    column[i0] = indbs[iflat]
                    i0 += 1

                    # non-diagonal elements (symmetric)
                    for jflat in overlapi:

                        if cropbs_flat is not False and not cropbs_flat[jflat]:
                            continue

                        jr = jflat % nx
                        jz = jflat // nx

                        # store (i, j) and (j, i) (symmetric matrix)
                        if jr >= ir:
                            iidR = idR[jr - ir, ir]
                            iiR = iR[jr - ir, ir]
                        else:
                            iidR = idR[abs(jr - ir), jr]
                            iiR = iR[abs(jr - ir), jr]
                        if jz >= iz:
                            iidZ = idZ[jz - iz, iz]
                            iiZ = iZ[jz - iz, iz]
                        else:
                            iidZ = idZ[abs(jz - iz), jz]
                            iiZ = iZ[abs(jz - iz), jz]
                        datadR[i0:i0+2] = iidR * iiZ
                        datadZ[i0:i0+2] = iiR * iidZ
                        row[i0:i0+2] = (indbs[iflat], indbs[jflat])
                        column[i0:i0+2] = (indbs[jflat], indbs[iflat])
                        i0 += 2

            assert i0 == nbtot
            opmat = (
                scpsp.csc_matrix((datadR, (row, column)), shape=shape),
                scpsp.csc_matrix((datadZ, (row, column)), shape=shape),
            )

    # ------------
    # D2N2

    elif operator == 'D2N2':

        # pre-compute integrals
        if deg in [0, 1]:
            msg = f"degree {deg} too low for operator {operator}"
            raise Exception(msg)
        if deg == 2:
            id2R = _operators._D2N2_Deg2(knotsx_mult, geom)
            id2Z = _operators._D2N2_Deg2(knotsy_mult)
            idR = _operators._D1N2_Deg2(knotsx_mult, geom)
            idZ = _operators._D1N2_Deg2(knotsy_mult)
            iR = _operators._D0N2_Deg2(knotsx_mult, geom)
            iZ = _operators._D0N2_Deg2(knotsy_mult)
        elif deg == 3:
            msg = "Integral D2N2 not implemented for deg=3!"
            raise NotImplementedError(msg)

        # set non-diagonal elements
        i0 = 0
        for ir in range(nx):
            for iz in range(ny):

                iflat = ir + iz*nx
                if cropbs_flat is not False and not cropbs_flat[iflat]:
                    continue

                # general case
                overlapi = overlap[:, iflat][overlap[:, iflat] > iflat]

                # diagonal element
                datad2R[i0] = id2R[0, ir] * iZ[0, iz]
                datad2Z[i0] = iR[0, ir] * id2Z[0, iz]
                datadRZ[i0] = idR[0, ir] * idZ[0, iz]
                row[i0] = indbs[iflat]
                column[i0] = indbs[iflat]
                i0 += 1

                # non-diagonal elements (symmetric)
                for jflat in overlapi:

                    if cropbs_flat is not False and not cropbs_flat[jflat]:
                        continue

                    jr = jflat % nx
                    jz = jflat // nx

                    # store (i, j) and (j, i) (symmetric matrix)
                    if jr >= ir:
                        iid2R = id2R[jr - ir, ir]
                        iidR = idR[jr - ir, ir]
                        iiR = iR[jr - ir, ir]
                    else:
                        iid2R = id2R[abs(jr - ir), jr]
                        iidR = idR[abs(jr - ir), jr]
                        iiR = iR[abs(jr - ir), jr]
                    if jz >= iz:
                        iid2Z = id2Z[jz - iz, iz]
                        iidZ = idZ[jz - iz, iz]
                        iiZ = iZ[jz - iz, iz]
                    else:
                        iid2Z = id2Z[abs(jz - iz), jz]
                        iidZ = idZ[abs(jz - iz), jz]
                        iiZ = iZ[abs(jz - iz), jz]
                    datad2R[i0:i0+2] = iid2R * iiZ
                    datad2Z[i0:i0+2] = iiR * iid2Z
                    datadRZ[i0:i0+2] = iidR * iidZ
                    row[i0:i0+2] = (indbs[iflat], indbs[jflat])
                    column[i0:i0+2] = (indbs[jflat], indbs[iflat])
                    i0 += 2

        assert i0 == nbtot
        opmat = (
            scpsp.csc_matrix((datad2R, (row, column)), shape=shape),
            scpsp.csc_matrix((datad2Z, (row, column)), shape=shape),
            scpsp.csc_matrix((datadRZ, (row, column)), shape=shape),
        )

    # ------------
    # D3N2

    elif operator == 'D3N2' and deg == 3:

        raise NotImplementedError("Integral D3N2 not implemented for deg=3!")

    return opmat, operator, geometry





import itertools as itt


import numpy as np
import scipy.interpolate as scpinterp


from . import _utils_bsplines


# #############################################################################
# #############################################################################
#                   class
# #############################################################################


# class BivariateSplinePolar():
    # """ Subclass for tofu

    # Defined from knots (unique) and deg
    # coefs set to 1 by default

    # Used self.set_coefs() to update
    # """

    # def __init__(self, knotsr=None, knotsa=None, deg=None):

        # assert deg in [0, 1, 2, 3], deg

        # # get knots pr bs
        # self._get_knots_per_bs_for_basis_elements(
            # knotsr=knotsr,
            # knotsa=knotsa,
            # deg=deg,
        # )

        # # get nbs
        # self.nbs = np.sum(self.nbs_a_per_r)

        # if self.knotsa is None:
            # self.shapebs = (self.nbs,)
        # elif np.unique(self.nbs_a_per_r).size == 1:
            # self.shapebs = (self.nbs_r, self.nbs_a_per_r[0])
        # else:
            # self.shapebs = (self.nbs,)

        # # func to get coef from (ii, jj)
        # if self.knotsa is not None:
            # self.func_coef_ind = self._get_func_coef_ind()

        # # deg
        # self.deg = deg

    # def _get_knots_per_bs_for_basis_elements(
        # self,
        # knotsr=None,
        # knotsa=None,
        # deg=None,
    # ):

        # # ------------------------------------
        # # get knots per bs in radius direction

        # knots_per_bs_r = _utils_bsplines._get_knots_per_bs(
            # knotsr, deg=deg, returnas='data',
        # )
        # knotsr_with_mult, nbsr = _utils_bsplines._get_knots_per_bs(
            # knotsr, deg=deg, returnas='data', return_unique=True,
        # )

        # if nbsr != knots_per_bs_r.shape[1]:
            # msg = "Inconsistent nb. of splines in r direction"
            # raise Exception(msg)

        # # ----------------
        # # check angle

        # if isinstance(knotsa, np.ndarray) and knotsa.ndim == 1:
            # knotsa = [knotsa for ii in range(nbsr)]

        # c0 = (
            # isinstance(knotsa, list)
            # and len(knotsa) == nbsr
            # and all([
                # aa is None
                # or (
                    # isinstance(aa, np.ndarray)
                    # and aa.ndim == 1
                    # and np.allclose(
                        # aa,
                        # np.unique(np.arctan2(np.sin(aa), np.cos(aa))),
                        # equal_nan=False,
                    # )
                # )
                # for aa in knotsa
            # ])
        # )
        # if knotsa is not None and not c0:
            # msg = (
                # f"Arg angle must a list of {nbsr} elements, where each can be:\n"
                # "\t- None: no poloidal discretization\n"
                # "\t- array: 1d array of sorted angles in radians in [-pi, pi[\n"
                # f"Provided: {knotsa}"
            # )
            # raise Exception(msg)

        # # make sure to remove double angles
        # # particularly at edges (-pi vs pi)
        # if knotsa is not None:
            # for ii, aa in enumerate(knotsa):
                # if aa is not None:
                    # aa = np.unique(np.arctan2(np.sin(aa), np.cos(aa)))
                    # damin = np.min(np.diff(aa))
                    # if np.abs(aa[-1] - (aa[0] + 2.*np.pi)) < damin/100.:
                        # aa = aa[:-1]
                    # knotsa[ii] = aa

        # # ----------------
        # # Pre-compute bsplines basis elements

        # lbr = [
            # scpinterp.BSpline.basis_element(
                # knots_per_bs_r[:, ii],
                # extrapolate=False,
            # )
            # for ii in range(nbsr)
        # ]

        # # angle bsplines
        # nbsa = np.ones((nbsr,), dtype=int)
        # if knotsa is not None:
            # lba = [[] for ii in range(nbsr)]
            # knots_per_bs_a = []
            # for ii in range(nbsr):

                # if knotsa[ii] is None:
                    # lba.append(None)
                    # knots_per_bs_a.append(None)
                    # continue

                # knots_per_bsai = _utils_bsplines._get_knots_per_bs(
                    # knotsa[ii],
                    # deg=deg,
                    # returnas='data',
                    # poloidal=True,
                # )

                # nbsa[ii] = knots_per_bsai.shape[1]
                # knots_per_bs_a.append(knots_per_bsai)

                # for jj in range(nbsa[ii]):
                    # kj = knots_per_bsai[:, jj]
                    # lba[ii].append(scpinterp.BSpline.basis_element(
                        # kj,
                        # extrapolate=False,
                    # ))
                    # if kj[0] >= kj[-1]:
                        # i2pi = np.r_[False, np.diff(kj) < 0].nonzero()[0][0]
                        # lba[ii][-1].t[deg + i2pi:] += 2.*np.pi
        # else:
            # lba = None
            # knots_per_bs_a = None

        # # ----------------
        # # bsplines centers

        # cents_per_bs_r = _utils_bsplines._get_cents_per_bs(
            # 0.5*(knotsr[1:] + knotsr[:-1]),
            # deg=deg,
            # returnas='data',
        # )
        # if knotsa is None:
            # cents_per_bs_a = None
        # else:
            # cents_per_bs_a = [None for ii in range(nbsr)]
            # for ii in range(nbsr):
                # if knotsa[ii] is None:
                    # pass
                # else:
                    # cents_per_bs_a[ii] = _utils_bsplines._get_cents_per_bs(
                        # 0.5*(knotsa[ii][1:] + knotsa[ii][:-1]),
                        # deg=deg,
                        # returnas='data',
                        # poloidal=True,
                    # )

        # # ----------------
        # # bsplines apex

        # apex_per_bs_r = _utils_bsplines._get_apex_per_bs(
            # knots=knotsr,
            # knots_per_bs=knots_per_bs_r,
            # deg=deg,
        # )
        # if knotsa is None:
            # apex_per_bs_a = None
        # else:
            # apex_per_bs_a = [None for ii in range(nbsr)]
            # for ii in range(nbsr):
                # if knotsa[ii] is None:
                    # pass
                # else:
                    # apex_per_bs_a[ii] = _utils_bsplines._get_apex_per_bs(
            # knots=knotsa[ii],
            # knots_per_bs=knots_per_bs_a[ii],
            # deg=deg,
            # poloidal=True,
        # )

        # # ------
        # # store

        # self.knotsr = knotsr
        # self.knotsa = knotsa
        # self.knotsr_with_mult = knotsr_with_mult
        # self.knotsa_with_mult = knotsa
        # self.knots_per_bs_r = knots_per_bs_r
        # self.knots_per_bs_a = knots_per_bs_a
        # self.cents_per_bs_r = cents_per_bs_r
        # self.cents_per_bs_a = cents_per_bs_a
        # self.apex_per_bs_r = apex_per_bs_r
        # self.apex_per_bs_a = apex_per_bs_a
        # self.nbs_r = nbsr
        # self.nbs_a_per_r = nbsa
        # self.lbr = lbr
        # self.lba = lba

        # # -----------------
        # # compute knots per bs for flattened

        # # if deg == 0:
            # # pass
        # # else:
            # # knots_per_bs_r = np.concatenate(
                # # (
                    # # np.tile(knots_per_bs_r[0, :] - 1, (deg, 1)),
                    # # knots_per_bs_r,
                    # # np.tile(knots_per_bs_r[-1, :] + 1, (deg, 1)),
                # # ),
                # # axis=0,
            # # )

            # # for ii in range(nbsr):
                # # if knotsa[ii] is not None:
                    # # knots_per_bs_a[ii] = np.concatenate(
                        # # (
                            # # np.tile(knots_per_bs_a[ii][0, :] - 1, (deg, 1)),
                            # # knots_per_bs_a[ii],
                            # # np.tile(knots_per_bs_a[ii][-1, :] + 1, (deg, 1)),
                        # # ),
                        # # axis=0,
                    # # )

        # # self.knots_per_bs_r_pad = np.asfortranarray(knots_per_bs_r)
        # # self.knots_per_bs_a_pad = np.asfortranarray(knots_per_bs_a)

    # def _get_func_coef_ind(self):
        # def func(ii, jj):
            # return np.r_[0, np.cumsum(self.nbs_a_per_r)][ii] + jj
        # return func

    # def _check_radiusangle_input(self, radius=None, angle=None):

        # if not isinstance(radius, np.ndarray):
            # radius = np.atleast_1d(radius)

        # if self.knotsa is not None:
            # err = False
            # if angle is not None:
                # if not isinstance(angle, np.ndarray):
                    # angle = np.atleast_1d(angle)
                # if angle.shape != radius.shape:
                    # err = True
            # else:
                # if len(self.shapebs) == 2:
                    # err = True
                # else:
                    # shape = np.r_[np.ones((radius.ndim,), dtype=int), -1]
                    # kpbrmin = self.knots_per_bs_r[0, :].reshape(shape)
                    # kpbrmax = self.knots_per_bs_r[-1, :].reshape(shape)
                    # indr = (
                        # (radius[..., None] >= kpbrmin)
                        # & (radius[..., None] <= kpbrmax)
                    # )
                    # for ii in range(radius.ndim):
                        # indr = np.any(indr, axis=0)

                    # if np.any((self.nbs_a_per_r > 1) & indr):
                        # err = True
            # if err:
                # msg = (
                    # "Arg angle must be a np.ndarray same shape as radius!\n"
                    # f"\t- radius.shape = {radius.shape}\n"
                    # f"\t- angle = {angle}\n"
                    # "It should be an array of poloidal angles in radians"
                # )
                # raise Exception(msg)

        # return radius, angle

    # def __call__(
        # self,
        # # coordiantes
        # radius=None,
        # angle=None,
        # # coefs
        # coefs=None,
        # axis=None,
        # # options
        # radius_vs_time=None,
        # axis_time=None,
        # val_out=None,
        # # for purely radial only
        # deriv=None,
        # # for compatibility (unused)
        # **kwdargs,
    # ):
        # """ Assumes

        # coefs.shape = (nt, shapebs)

        # """

        # # ------------
        # # check inputs

        # if deriv is None:
            # deriv = 0
        # if self.knotsa is not None and deriv != 0:
            # msg = "Derivatives not-implemented for non-purely radial bsplines!"
            # raise Exception(msg)
        # if radius_vs_time:
            # assert isinstance(axis_time, int)

        # # coefs
        # self._check_coefs(coefs=coefs)

        # radius, angle = self._check_radiusangle_input(
            # radius=radius,
            # angle=angle,
        # )

        # # coefs
        # (
            # shape_ra, shape_other, axis_ra, ind_coefs, ind_ra,
        # ) = _utils_bsplines._coefs_axis(
            # shapebs=self.shapebs,
            # coefs=coefs,
            # axis=axis,
            # shapex=radius.shape,
        # )
        # shape_coefs = coefs.shape
        # val = np.zeros(shape_ra, dtype=float)

        # # ------------
        # # compute

        # if self.knotsa is None:

            # if radius_vs_time:
                # for ind in itt.product(*[range(aa) for aa in shape_other]):

                    # sli_c, sli_x = _utils_bsplines._get_slices_iter(
                        # axis=axis,
                        # axis_x=axis_ra,
                        # shape_coefs=shape_coefs,
                        # shape_x=shape_ra,
                        # ind=ind,
                        # ind_coefs=ind_coefs,
                        # ind_x=ind_ra,
                    # )

                    # val[sli_x] = scpinterp.BSpline(
                        # self.knotsr_with_mult,
                        # coefs[sli_c],
                        # self.deg,
                        # extrapolate=False,
                    # )(radius[sli_x], nu=deriv)

            # else:
                # for ind in itt.product(*[range(aa) for aa in shape_other]):

                    # sli_c, sli_x = _utils_bsplines._get_slices_iter(
                        # axis=axis,
                        # axis_x=axis_ra,
                        # shape_coefs=shape_coefs,
                        # shape_x=shape_ra,
                        # ind=ind,
                        # ind_coefs=ind_coefs,
                        # ind_x=ind_ra,
                    # )

                    # val[sli_x] = scpinterp.BSpline(
                        # self.knotsr_with_mult,
                        # coefs[sli_c],
                        # self.deg,
                        # extrapolate=False,
                    # )(radius, nu=deriv)




            # if radius_vs_time:
                # for it in range(nt):
                    # val[it, ...] = scpinterp.BSpline(
                        # self.knotsr_with_mult,
                        # coefs[it, :],
                        # self.deg,
                        # extrapolate=False,
                    # )(radius[it, ...], nu=deriv)
            # else:
                # for it in range(nt):
                    # val[it, ...] = scpinterp.BSpline(
                        # self.knotsr_with_mult,
                        # coefs[it, :],
                        # self.deg,
                        # extrapolate=False,
                    # )(radius, nu=deriv)

        # elif radius_vs_time:

            # for it in range(nt):
                # for ii, nbsa in enumerate(self.nbs_a_per_r):
                    # iok = (
                        # (radius[it, ...] >= self.knots_per_bs_r[0, ii])
                        # & ((radius[it, ...] < self.knots_per_bs_r[-1, ii]))
                    # )
                    # valr = self.lbr[ii](radius[it, iok])
                    # if nbsa == 1:
                        # ind = self.func_coef_ind(ii, 0)
                        # val[it, iok] += coefs[it, ind] * valr
                    # else:
                        # for jj in range(nbsa):
                            # kj = self.knots_per_bs_a[ii][:, jj]
                            # if kj[0] > kj[-1]:
                                # atemp = np.copy(angle[it, iok])
                                # atemp[atemp < kj[0]] += 2.*np.pi
                                # vala = self.lba[ii][jj](atemp)
                            # else:
                                # vala = self.lba[ii][jj](angle[it, iok])

                            # iokj = ~np.isnan(vala)
                            # if np.any(iokj):
                                # iok2 = np.copy(iok)
                                # iok2[iok2] = iokj
                            # else:
                                # continue

                            # if len(self.shapebs) == 1:
                                # ind = self.func_coef_ind(ii, jj)
                                # val[it, iok2] += (
                                    # coefs[it, ind] * (valr[iokj]*vala[iokj])
                                # )
                            # else:
                                # val[it, iok2] += (
                                    # coefs[it, ii, jj] * (valr[iokj]*vala[iokj])
                                # )

        # else:
            # for ii, nbsa in enumerate(self.nbs_a_per_r):
                # iok = (
                    # (radius >= self.knots_per_bs_r[0, ii])
                    # & ((radius < self.knots_per_bs_r[-1, ii]))
                # )
                # valr = self.lbr[ii](radius[iok])
                # if nbsa == 1:
                    # ind = self.func_coef_ind(ii, 0)
                    # val[:, iok] += coefs[:, ind][:, None] * valr[None, ...]
                # else:
                    # for jj in range(nbsa):
                        # kj = self.knots_per_bs_a[ii][:, jj]
                        # if kj[0] > kj[-1]:
                            # atemp = np.copy(angle[iok])
                            # atemp[atemp < kj[0]] += 2.*np.pi
                            # vala = self.lba[ii][jj](atemp)
                        # else:
                            # vala = self.lba[ii][jj](angle[iok])

                        # iokj = np.isfinite(vala)
                        # if np.any(iokj):
                            # iok2 = np.copy(iok)
                            # iok2[iok2] = iokj
                        # else:
                            # continue

                        # if len(self.shapebs) == 1:
                            # ind = self.func_coef_ind(ii, jj)
                            # val[:, iok2] += (
                                # coefs[:, ind][:, None]
                                # * (valr[iokj]*vala[iokj])[None, ...]
                            # )
                        # else:
                            # val[:, iok2] += (
                                # coefs[:, ii, jj][:, None]
                                # * (valr[iokj]*vala[iokj])[None, ...]
                            # )

        # # clean out-of-mesh
        # if val_out is not False:
            # # pts out
            # indout = (
                # (radius < self.knotsr.min())
                # | (radius > self.knotsr.max())
            # )
            # if radius_vs_time:
                # val[indout] = val_out
            # else:
                # val[:, indout] = val_out

        # return val

    # def ev_details(
        # self,
        # # coordiantes
        # radius=None,
        # angle=None,
        # # options
        # indbs_tf=None,
        # deriv=None,
        # # for compatibility (unused)
        # **kwdargs,
    # ):
        # """ Assumes

        # coefs.shape = (nt, shapebs)
        # indbs_tf = flat array of int indices

        # """

        # # ------------
        # # check inputs

        # radius, angle = self._check_radiusangle_input(
            # radius=radius,
            # angle=angle,
        # )

        # if indbs_tf is None:
            # nbs = self.nbs
        # else:
            # indbs_tf = np.atleast_1d(indbs_tf).ravel()
            # assert 'int' in indbs_tf.dtype.name, indbs_tf
            # assert np.unique(indbs_tf).size == indbs_tf.size
            # nbs = indbs_tf.size

        # if deriv is None:
            # deriv = 0

        # # --------
        # # prepare

        # shape = tuple(np.r_[radius.shape, nbs])
        # val = np.zeros(shape)

        # # ------------
        # # compute

        # if self.knotsa is None:

            # # radius only
            # ni = 0
            # for ii in range(self.nbs_r):

                # if indbs_tf is not None and ii not in indbs_tf:
                    # continue

                # iok = (
                    # (radius >= self.knots_per_bs_r[0, ii])
                    # & ((radius < self.knots_per_bs_r[-1, ii]))
                # )
                # if np.any(iok):
                    # val[iok, ni] = self.lbr[ii](radius[iok], nu=deriv)
                # ni += 1

        # else:

            # # if np.abs(np.nanmin(radius) - 0.1800309517775584) < 0.0001:
                # # import pdb.set_trace()      # DB  # chan = 27

            # # radius + angle
            # ni = 0
            # for ii, nbsa in enumerate(self.nbs_a_per_r):

                # c0 = (
                    # nbsa == 1
                    # and indbs_tf is not None
                    # and self.func_coef_ind(ii, 0) not in indbs_tf
                # )
                # if c0:
                    # continue

                # # compute valr
                # iok = (
                    # (radius >= self.knots_per_bs_r[0, ii])
                    # & ((radius < self.knots_per_bs_r[-1, ii]))
                # )

                # if not np.any(iok):
                    # ni += nbsa
                    # continue

                # valr = self.lbr[ii](radius[iok], nu=deriv)

                # # compute vala
                # if nbsa == 1:
                    # val[iok, ni] = valr
                    # ni += 1
                # else:
                    # if deriv > 0:
                        # msg = "Derivative > 0 not implemented for 2d splines"
                        # raise Exception(msg)

                    # for jj in range(nbsa):
                        # ind = self.func_coef_ind(ii, jj)

                        # if indbs_tf is not None and ind not in indbs_tf:
                            # continue

                        # kj = self.knots_per_bs_a[ii][:, jj]
                        # if kj[0] >= kj[-1]:
                            # atemp = np.copy(angle[iok])
                            # atemp[atemp < kj[0]] += 2.*np.pi
                            # vala = self.lba[ii][jj](atemp)
                        # else:
                            # vala = self.lba[ii][jj](angle[iok])

                        # iokj = np.isfinite(vala)
                        # if np.any(iokj):
                            # iok2 = np.copy(iok)
                            # iok2[iok2] = iokj
                            # val[iok2, ni] = valr[iokj]*vala[iokj]

                        # # if ii == 4 and jj == 3:
                            # # import pdb; pdb.set_trace()         # DB
                        # ni += 1

        # return val



# -*- coding: utf-8 -*-


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import datastock as ds


from . import _class02_interpolate_all as _interpolate_all
from . import _class02_plot_as_profile2d as _plot_as_profile2d


# ##############################################################
# ##############################################################
#                           Main
# ##############################################################


def plot_as_profile2d_compare(
    # ressources
    coll=None,
    # inputs
    keys=None,
    # parameters
    dres=None,
    # contours
    dlevels=None,
    ref_com=None,
    # details
    plot_details=None,
    # ref vectors
    dref_vectorZ=None,
    dref_vectorU=None,
    # interpolation
    val_out=None,
    nan0=None,
    # plotting
    vmin=None,
    vmax=None,
    cmap=None,
    vmin_err=None,
    vmax_err=None,
    cmap_err=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    # interactivity
    dinc=None,
    connect=None,
):

    # --------------
    # check input

    if connect is None:
        connect = True

    (
        dkeys,
        dlevels,
        cmap, cmap_err,
        vmin, vmax,
        dcolorbar, dleg,
        connect,
    ) = _plot_as_profile2d._check(
        coll=coll,
        keys=keys,
        dlevels=dlevels,
        # plotting
        cmap=cmap,
        cmap_err=cmap_err,
        # figure
        dcolorbar=dcolorbar,
        dleg=dleg,
        # interactivity
        connect=connect,
    )

    # ---------------
    # prepare dax

    if dax is None:
        dax = _create_axes(
            dkeys=dkeys,
            fs=fs,
            dmargin=dmargin,
        )

    lk0 = ['prof0', 'vert', 'hor', 'traces', 'spectrum', 'radial']
    lk1 = ['prof1', 'vert', 'hor', 'traces', 'spectrum', 'radial']
    dax2 = {
        keys[0]: {k0: dax[k0] for k0 in lk0 if k0 in dax.keys()},
        keys[1]: {k0: dax[k0] for k0 in lk1 if k0 in dax.keys()},
    }

    # ----------------------
    # get unique 2d sampling

    keys, _, dres, submesh = _interpolate_all._check(
        coll=coll,
        keys=keys,
        # sampling
        dres=dres,
        submesh=True,
    )

    # list meshes 2d
    wm = coll._which_mesh
    lm = list(dres.keys())
    lm = [k0 for k0 in lm if coll.dobj[wm][k0]['nd'] == '2d']

    if dres[lm[0]]['x0'] is None:
        dunique = coll.get_sample_mesh(
            key=lm[0],
            res=dres[lm[0]]['res'],
            mode=dres[lm[0]]['mode'],
            grid=False,
            # store
            store=False,
            kx0='unique_x0',
            kx1='unique_x1',
        )
    else:
        raise NotImplementedError()
        dunique = {
            'x0': {},
            'x1': {},
        }

    # ---------------
    # plot profiles2d

    collax, dgroup = coll.plot_as_profile2d(
        key=keys,
        dres=dres,
        dunique_mesh_2d=dunique,
        dlevels=dlevels,
        ref_com=ref_com,
        # details
        plot_details=plot_details,
        # ref vectors
        dref_vectorZ=dref_vectorZ,
        dref_vectorU=dref_vectorU,
        # interpolation
        val_out=val_out,
        nan0=nan0,
        # plotting
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        # figure
        dax=dax2,
        dcolorbar=None,
        # interactivity
        connect=False,
    )

    # ----------
    # plot error

    # -----------
    # connect

    if connect is True:
        collax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
        collax.disconnect_old()
        collax.connect()

        collax.show_commands()
        return collax
    else:
        return collax, dgroup


# ################################################################
# ################################################################
#                       Create axes
# ################################################################


def _create_axes(
    dkeys=None,
    fs=None,
    dmargin=None,
):

    # ------------------
    # check and prepare

    ndim_extra = len(list(dkeys.values())[0]['ref_other'])
    hassubmesh = any([v0['submesh'] is not None for v0 in dkeys.values()])
    if hassubmesh:
        ndim_extra += 1

    if fs is None:
        fs = (15, 9)

    if dmargin is None:
        dmargin = {
            'left': 0.04, 'right': 0.98,
            'bottom': 0.06, 'top': 0.92,
            'hspace': 0.4, 'wspace': 2.,
        }

    # ----------------
    # create axis grid

    # axes for images
    dgs = {}
    nrows = 3
    if ndim_extra == 0:
        ncols = 17
    else:
        ncols = 22

    gs = gridspec.GridSpec(ncols=ncols, nrows=nrows, **dmargin)

    dgs['prof0'] = gs[:2, -17:-13]
    dgs['prof1'] = gs[:2, -13:-9]
    dgs['vert'] = gs[:2, -9:-7]
    dgs['err'] = gs[:2, -6:-2]
    dgs['vert_err'] = gs[:2, -2:]
    dgs['hor'] = gs[2, -17:-13]
    dgs['hor_err'] = gs[2, -6:-2]

    for ii in range(0, ndim_extra):
        if ii == 0 and hassubmesh:
            kk = 'radial'
        else:
            kk = ['traces', 'spectrum'][ii - hassubmesh]
        dgs[kk] = gs[ii, :4]

    # ----------------
    # prepare figure and dax

    fig = plt.figure(figsize=fs)

    dax = {}

    # ------------
    # 2d profiles

    kax = 'prof0'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(dgs[kax])
        dax[kax] = {'handle': ax, 'type': 'matrix'}

    kax = 'prof1'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharex=dax['prof0']['handle'],
            sharey=dax['prof0']['handle'],
        )
        dax[kax] = {'handle': ax, 'type': 'matrix'}

    kax = 'err'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharex=dax['prof0']['handle'],
            sharey=dax['prof0']['handle'],
        )
        dax[kax] = {'handle': ax, 'type': 'matrix'}

        ax.set_title('difference', size=12, fontweight='bold')

    # --------------------
    # hor and vert slices

    kax = 'hor'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharex=dax['prof0']['handle'],
        )
        dax[kax] = {'handle': ax, 'type': 'horizontal'}

    kax = 'hor_err'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharex=dax['prof0']['handle'],
        )
        dax[kax] = {'handle': ax, 'type': 'horizontal'}

    kax = 'vert'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharey=dax['prof0']['handle'],
        )
        dax[kax] = {'handle': ax, 'type': 'vertical'}

    kax = 'vert_err'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharey=dax['prof0']['handle'],
        )
        dax[kax] = {'handle': ax, 'type': 'vertical'}

    # --------------------
    # extra dimensions

    kax = 'traces'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharey=dax['hor']['handle'],
        )
        dax[kax] = {'handle': ax, 'type': 'tracesZ'}

    kax = 'spectrum'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharey=dax['hor']['handle'],
        )
        dax[kax] = {'handle': ax, 'type': 'tracesU'}

    kax = 'radial'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharey=dax['hor']['handle'],
        )
        dax[kax] = {'handle': ax}

    kax = 'radial_err'
    if dgs.get(kax) is not None:
        ax = fig.add_subplot(
            dgs[kax],
            sharex=dax['radial']['handle'],
            sharey=dax['hor']['handle'],
        )
        dax[kax] = {'handle': ax}

    return dax





# #####################
# Debugging unit tests - START
# #####################

import sys
import os
_PATH_HERE = os.path.dirname(__file__)
_PATH_PROJ = os.path.dirname(os.path.dirname(_PATH_HERE))
_PATH_DS = os.path.join(_PATH_PROJ, 'datastock')
sys.path.insert(0, _PATH_DS)
import datastock as ds
sys.path.pop(0)

# ######################
# Debugging unit tests - END
# ######################


from ._class03_Bins import Bins as BSplines2D
from . import tests



# -*- coding: utf-8 -*-


# common
import matplotlib.pyplot as plt


_LALLOWED_AXESTYPES = [
    None,
    'cross',
    'hor',
    '3d',
    'camera',
    'matrix',
    'timetrace',
    'profile1d',
    'image',
    'text',
    'misc',
]


# ##################################################################
# ##################################################################
#                   Utilities for plotting
# ##################################################################


def _check_dax(dax=None, main=None):

    # None
    if dax is None:
        return dax

    # Axes
    if issubclass(dax.__class__, plt.Axes):
        if main is None:
            msg = (
            )
            raise Exception(msg)
        else:
            return {main: dax}

    # dict
    c0 = (
        isinstance(dax, dict)
        and all([
            isinstance(k0, str)
            and (
                (
                    issubclass(v0.__class__, plt.Axes)
                )
                or (
                    isinstance(v0, dict)
                    and issubclass(v0.get('handle').__class__, plt.Axes)
                    # and v0.get('type') in _LALLOWED_AXESTYPES
                )
            )
            for k0, v0 in dax.items()
        ])
    )
    if not c0:
        msg = (
            "Arg dax is not conform:\n"
            + str(dax)
        )
        import pdb; pdb.set_trace()     # DB
        raise Exception(msg)

    for k0, v0 in dax.items():
        if issubclass(v0.__class__, plt.Axes):
            dax[k0] = {'handle': v0, 'type': k0}
        if isinstance(v0, dict):
            dax[k0]['type'] = v0.get('type')

    return dax



# -*- coding: utf-8 -*-


# Common
import numpy as np
import datastock as ds


# specific
from . import _generic_mesh


# #############################################################################
# #############################################################################
#                           mesh generic check
# #############################################################################


def check(
    coll=None,
    key=None,
    # knots
    knots=None,
    uniform=None,
    # defined from pre-existing bsplines
    subkey=None,
    # additional attributes
    **kwdargs,
):

    # --------
    # keys

    # key
    key = ds._generic_check._obj_key(
        d0=coll._dobj.get(coll._which_mesh, {}),
        short='m',
        key=key,
    )

    # ------------
    # knots vector

    knots, res = _check_knots(
        key=key,
        knots=knots,
        uniform=uniform,
    )

    # ----------------
    # angles handdling

    isangle = str(kwdargs.get('units')) == 'rad'
    if isangle:
        knots, cents = _knots_angle(knots)

    else:
        cents = 0.5*(knots[1:] + knots[:-1])

    # ------------------------
    # depend on other bsplines

    submesh, subbs, kwdargs = _defined_from(
        coll=coll,
        subkey=subkey,
        # parameters
        kwdargs=kwdargs,
    )

    # --------------
    # to dict

    dref, ddata, dobj = _to_dict(
        coll=coll,
        key=key,
        knots=knots,
        cents=cents,
        res=res,
        # sub quantity
        subkey=subkey,
        subbs=subbs,
        submesh=submesh,
        # attributes
        **kwdargs,
    )

    return key, dref, ddata, dobj


# ##################################################################
# ##################################################################
#                       knots
# ##################################################################


def _check_knots(
    key=None,
    knots=None,
    uniform=None,
):

    # ------------
    # check input

    uniform = ds._generic_check._check_var(
        uniform, 'uniform',
        types=bool,
        default=False,
    )

    # ---------
    # check x

    knots = ds._generic_check._check_flat1darray(
        knots, 'knots',
        dtype=float,
        unique=True,
        can_be_None=False,
    )

    # resolution
    res = np.diff(knots)

    # -----------------
    # check uniformity

    if np.allclose(res, np.mean(res), atol=1e-14, rtol=1e-6):
        res = res[0]

    elif uniform is True:
        msg = (
            f"Non-uniform resolution for user-provided mesh '{key}'\n"
            f"\t- unique res: {np.unique(res)}\n"
            f"\t- diff res: {np.diff(np.unique(res))}\n"
            f"\t- res: {res}\n"
            )
        raise NotImplementedError(msg)

    return knots, res


def _knots_angle(
    knots=None,
    res=None,
):

    # knots in ]-pi; pi]
    knots_temp = np.unique(np.arctan2(np.sin(knots), np.cos(knots)))
    if not np.allclose(knots, knots_temp):
        msg = (
            "Angle knots must be in ]-pi; pi]!\n"
            f"Provided: {knots}"
        )
        raise Exception(msg)

    # cents - handle discontinuity at -pi
    cents = 0.5*(knots[1:] + knots[:-1])
    mid = 0.5*(knots[-1] + (2.*np.pi + knots[0]))
    mid = np.arctan2(np.sin(mid), np.cos(mid))
    if mid < cents[0]:
        cents = np.r_[mid, cents]
    else:
        cents = np.r_[cents, mid]

    return knots, cents


# #############################################################################
# #############################################################################
#                        defined_from
# #############################################################################


def _defined_from(
    coll=None,
    subkey=None,
    nd=None,
    # parameters
    kwdargs=None,
):

    # ------------
    # trivial

    if subkey is None:
        return None, None, kwdargs

    # ------------
    # check key_on

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    if coll.dobj.get(wbs) is not None:
        dbs, dref = coll.get_dict_bsplines()
        lok =[
            k0 for k0, v0 in dbs.items()
            if len(v0) == 1
        ]

        if nd is not None:
            lok2 = []
            for k0 in lok:
                kbs = list(dbs[k0].keys())[0]
                if coll.dobj[wm][coll.dobj[wbs][kbs]['mesh']]['nd'] == nd:
                    lok2.append(k0)

            lok = lok2

    else:
        lok = []

    subkey = ds._generic_check._check_var(
        subkey, 'subkey',
        types=str,
        allowed=lok,
    )

    # ----------------
    # complete kwdargs

    lq = ['dim', 'quant', 'name', 'units']
    for k0 in lq:
        if kwdargs.get(k0) is None:
            kwdargs[k0] = str(coll.ddata[subkey][k0])

    # --------------
    # key_submesh

    subbs = coll.ddata[subkey][wbs][0]
    submesh = coll.dobj[wbs][subbs]['mesh']

    return submesh, subbs, kwdargs


# #############################################################################
# #############################################################################
#                           to_dict
# #############################################################################


def _to_dict(
    coll=None,
    key=None,
    knots=None,
    cents=None,
    res=None,
    # submesh
    subkey=None,
    subbs=None,
    submesh=None,
    # attributes
    **kwdargs,
):

    # ---------
    # prepare

    # keys
    # keys of knots and cents
    kkr, kcr, kk, kc = _generic_mesh.names_knots_cents(key=key)

    # variable
    variable = not np.isscalar(res)

    # attributes
    latt = ['dim', 'quant', 'name', 'units']
    dim, quant, name, units = [kwdargs.get(ss) for ss in latt]

    # subkey
    if subkey is not None:
        subkey = (subkey,)
    else:
        subkey = None

    # -------------
    # prepare dict

    # dref
    dref = {
        kkr: {
            'size': knots.size,
        },
        kcr: {
            'size': cents.size,
        },
    }

    # ddata
    ddata = {
        kk: {
            'data': knots,
            'units': units,
            # 'source': None,
            'dim': dim,
            'quant': quant,
            'name': name,
            'ref': kkr,
        },
        kc: {
            'data': cents,
            'units': units,
            # 'source': None,
            'dim': dim,
            'quant': quant,
            'name': name,
            'ref': kcr,
        },
    }

    # dobj
    dobj = {
        coll._which_mesh: {
            key: {
                'nd': '1d',
                'type': None,
                'knots': (kk,),
                'cents': (kc,),
                'shape-c': (cents.size,),
                'shape-k': (knots.size,),
                'variable': variable,
                'subkey': subkey,
                'subbs': subbs,
                'submesh': submesh,
                'crop': False,
            },
        },
    }

    # additional attributes
    for k0, v0 in kwdargs.items():
        if k0 not in latt:
            dobj[coll._which_mesh][key][k0] = v0

    return dref, ddata, dobj



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import itertools as itt


# Standard
import numpy as np


# specific
from . import test_input
from .. import _class01_checks_2d_rect as _checks
from .._class03_Bins import Bins as BSplines2D


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    pass


def teardown_module():
    pass


#######################################################
#
#     checking routines
#
#######################################################


class Test00_check_routines():

    def test00_mesh2DRect_X_check(self):

        lx = [[1, 2], [1, 2, 3, 4]]
        lres = [None, 10, 0.1, [0.1, 0.2], [0.1, 0.2, 0.3, 0.1]]

        for comb in itt.product(lx, lres):
            if hasattr(lres, '__iter__') and len(lres) != len(lx):
                continue
            x, res, ind = _checks._mesh2DRect_X_check(
                x=[1, 2, 3, 4],
                res=10,
            )
            if hasattr(lres, '__iter__'):
                assert x.size == np.unique(x).size == res.size + 1


#######################################################
#
#     Fixed meshes (1d, rect and tri, no submesh)
#
#######################################################


class Test01_BSplines2D():

    @classmethod
    def setup_class(cls):
        cls.bs = BSplines2D()

    @classmethod
    def setup_method(self):
        pass

    @classmethod
    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    # #################
    #  mesh 1d (fixed)
    # #################

    def test01_mesh1d_from_knots_uniform(self):
        test_input._add_1d_knots_uniform(self.bs)

    def test02_mesh1d_from_knots_variable(self):
        test_input._add_1d_knots_variable(self.bs)

    # ##############
    #  mesh 2d rect
    # ##############

    def test03_add_mesh_rect_uniform(self):
        test_input._add_rect_uniform(self.bs)

    def test04_add_mesh_rect_variable(self):
        test_input._add_rect_variable(self.bs)

    def test05_add_mesh_rect_variable_crop(self):
        test_input._add_rect_variable_crop(self.bs)

    def test06_add_mesh_rect_variable_crop(self):
        test_input._add_rect_variable_crop(self.bs)

    def test07_add_mesh_rect_variable_crop(self):
        test_input._add_rect_variable_crop(self.bs)

    # ##############
    #  mesh 2d tri
    # ##############

    def test08_add_mesh_tri_ntri1(self):
        test_input._add_tri_ntri1(self.bs)

    def test09_add_mesh_tri_ntri2(self):
        test_input._add_tri_ntri2(self.bs)

    def test10_add_mesh_tri_delaunay(self):
        test_input._add_tri_delaunay(self.bs)

    # ##############
    #  select
    # ##############

    def test11_select_mesh_element_1d(self):
        test_input._select_mesh_elements(self.bs, nd='1d', kind=None)

    def test12_select_mesh_element_rect(self):
        test_input._select_mesh_elements(self.bs, nd='2d', kind='rect')

    def test13_select_mesh_element_tri(self):
        test_input._select_mesh_elements(self.bs, nd='2d', kind='tri')

    # ##############
    #  sample
    # ##############

    def test14_sample_mesh_1d(self):
        test_input._sample_mesh(self.bs, nd='1d', kind=None)

    def test15_sample_mesh_rect(self):
        test_input._sample_mesh(self.bs, nd='2d', kind='rect')

    def test16_sample_mesh_tri(self):
        test_input._sample_mesh(self.bs, nd='2d', kind='tri')

    # ##############
    #  plot
    # ##############

    def test17_plot_mesh_1d(self):
        test_input._plot_mesh(self.bs, nd='1d', kind=None)

    def test18_plot_mesh_rect(self):
        test_input._plot_mesh(self.bs, nd='2d', kind='rect')

    def test19_plot_mesh_tri(self):
        test_input._plot_mesh(self.bs, nd='2d', kind='tri')

    # ##############
    #  add bsplines
    # ##############

    def test20_add_bsplines_1d(self):
        test_input._add_bsplines(self.bs, nd='1d')

    def test21_add_bsplines_2d_rect(self):
        test_input._add_bsplines(self.bs, kind='rect')

    def test22_add_bsplines_2d_tri(self):
        test_input._add_bsplines(self.bs, kind='tri')

    # ##############
    #  select bsplines
    # ##############

    def test23_select_bsplines_1d(self):
        test_input._select_bsplines(self.bs, nd='1d', kind=None)

    def test24_select_bsplines_2d_rect(self):
        test_input._select_bsplines(self.bs, nd='2d', kind='rect')

    def test25_select_bsplines_2d_tri(self):
        test_input._select_bsplines(self.bs, nd='2d', kind='tri')

    # ###############
    #  add data vs bs
    # ###############

    def test26_add_data_1bs_fix_1d(self, remove=True):
        test_input._add_data_1bs_fix(self.bs, nd='1d', kind=None, remove=remove)

    def test27_add_data_1bs_fix_2d_rect(self, remove=True):
        test_input._add_data_1bs_fix(self.bs, nd='2d', kind='rect', remove=remove)

    def test28_add_data_1bs_fix_2d_tri(self, remove=True):
        test_input._add_data_1bs_fix(self.bs, nd='2d', kind='tri', remove=remove)

    def test29_add_data_1bs_arrays_1d(self, remove=False):
        test_input._add_data_1bs_arrays(self.bs, nd='1d', kind=None, remove=remove)

    def test30_add_data_1bs_arrays_2d_rect(self, remove=False):
        test_input._add_data_1bs_arrays(self.bs, nd='2d', kind='rect', remove=remove)

    def test31_add_data_1bs_arrays_2d_tri(self, remove=False):
        test_input._add_data_1bs_arrays(self.bs, nd='2d', kind='tri', remove=remove)

    def test32_add_data_multibs_arrays(self, remove=False):
        test_input._add_data_multibs_arrays(self.bs, remove=remove)

    # ##############
    # interp bs
    # ##############

    def test33_interpolate_bsplines_1d(self):
        test_input._interpolate(self.bs, nd='1d', kind=None, details=False)

    def test34_interpolate_bsplines_1d_details(self):
        test_input._interpolate(self.bs, nd='1d', kind=None, details=True)

    def test35_interpolate_bsplines_2d_rect(self):
        test_input._interpolate(self.bs, nd='2d', kind='rect', details=False)

    def test36_interpolate_bsplines_2d_rect_details(self):
        test_input._interpolate(self.bs, nd='2d', kind='rect', details=True)

    def test37_interpolate_bsplines_2d_tri(self):
        test_input._interpolate(self.bs, nd='2d', kind='tri', details=False)

    def test38_interpolate_bsplines_2d_tri_details(self):
        test_input._interpolate(self.bs, nd='2d', kind='tri', details=True)

    # ##############
    # binning 1d
    # ##############

    def test39_binning_1d(self):
        test_input._bin_bs(self.bs, nd='1d', kind=None)

    # ##############
    # plot bsplines
    # ##############

    def test40_plot_bsplines_1d(self):
        pass

    def test41_plot_bsplines_2d_rect(self):
        pass

    def test42_plot_bsplines_2d_tri(self):
        pass

    # ####################
    # add mesh with subkey
    # ####################

    def test43_add_mesh_1d_subkey_1d(self):
        test_input._add_mesh_1d_subkey(self.bs, nd='1d', kind=None)

    def test44_add_mesh_1d_subkey_rect(self):
        test_input._add_mesh_1d_subkey(self.bs, nd='2d', kind='rect')

    def test45_add_mesh_1d_subkey_tri(self):
        test_input._add_mesh_1d_subkey(self.bs, nd='2d', kind='tri')

    def test46_add_mesh_2d_rect_subkey_rect(self):
        test_input._add_mesh_2d_rect_subkey(self.bs, nd='2d', kind='rect')

    def test47_add_mesh_2d_rect_subkey_tri(self):
        test_input._add_mesh_2d_rect_subkey(self.bs, nd='2d', kind='tri')

    def test48_add_mesh_2d_rect_var_subkey_rect(self):
        test_input._add_mesh_2d_rect_subkey(self.bs, nd='2d', kind='rect')

    # ################################
    # add bsplines on mesh with subkey
    # ################################

    def test49_add_bsplines_subkey(self):
        test_input._add_bsplines(self.bs, subkey=True)

    # ################################
    # add data on bsplines with subkey
    # ################################

    def test50_add_data_subkey(self):
        test_input._add_data_multibs_arrays(
            self.bs,
            nd=None,
            kind=None,
            subbs=True,
            remove=False,
        )

    # ################################
    # interpolate data with subkey
    # ################################

    def test51_interpolate_subkey_1d(self):
        test_input._interpolate(
            self.bs,
            nd='1d',
            kind=None,
            submesh=True,
        )

    # def test48_interpolate_data_subkey_from_subkey(self):
        # test_input._interpolate_from_subkey(
            # self.bs,
        # )

    # ################################
    # interpolate data with subkey
    # ################################

    def test53_plot_as_profile2d(self):
        test_input._plot_as_profile2d(
            self.bs,
            nd='2d',
            kind=None,
        )

    # ################################
    # operators
    # ################################

    def test54_operators_1d(self):
        test_input._get_operators(
            self.bs,
            nd='1d',
            kind=None,
        )

    def test55_operators_2d_rect(self):
        test_input._get_operators(
            self.bs,
            nd='2d',
            kind='rect',
        )

    def test56_operators_2d_tri(self):
        pass

    def test57_operators_1d_subkey(self):
        pass



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
from matplotlib.path import Path
import datastock as ds


# specific
from . import _class01_select as _select


# #############################################################################
# #############################################################################
#                          crop rect mesh
# #############################################################################


def crop(
    coll=None,
    key=None,
    crop=None,
    thresh_in=None,
    remove_isolated=None,
):
    """ Crop a rect mesh

    Parameters
    ----------
    key:        str
        key of the rect mesh to be cropped
    crop:      np.ndarray
        Can be either:
            - bool: a boolean mask array
            - float: a closed 2d polygon used for cropping
    threshin:   int
        minimum nb. of corners for a mesh element to be included
    remove_isolated: bool
        flag indicating whether to remove isolated mesh elements

    Return
    ------
    crop:       np.ndarray
        bool mask
    key:        str
        key of the rect mesh to be cropped
    thresh_in:  int
        minimum nb. of corners for a mesh element to be included

    """

    # ------------
    # check inputs

    key, cropbool, thresh_in, remove_isolated = _crop_check(
        coll=coll, key=key, crop=crop, thresh_in=thresh_in,
        remove_isolated=remove_isolated,
    )

    # -----------
    # if crop is a poly => compute as bool

    if not cropbool:

        (Rc, Zc), (Rk, Zk) = coll.select_mesh_elements(
            key=key,
            elements='cents',
            return_neighbours=True,
            returnas='data',
        )

        nR, nZ = Rc.shape
        npts = Rk.shape[-1] + 1

        pts = np.concatenate(
            (
                np.concatenate((Rc[:, :, None], Rk), axis=-1)[..., None],
                np.concatenate((Zc[:, :, None], Zk), axis=-1)[..., None],
            ),
            axis=-1,
        ).reshape((npts*nR*nZ, 2))

        isin = Path(crop.T).contains_points(pts).reshape((nR, nZ, npts))
        crop = np.sum(isin, axis=-1) >= thresh_in

        # Remove isolated pixelsi
        if remove_isolated is True:
            # All pixels should have at least one neighbour in R and one in Z
            # This constraint is useful for discrete gradient evaluation (D1N2)
            neighR = np.copy(crop)
            neighR[0, :] &= neighR[1, :]
            neighR[-1, :] &= neighR[-2, :]
            neighR[1:-1, :] &= (neighR[:-2, :] | neighR[2:, :])
            neighZ = np.copy(crop)
            neighZ[:, 0] &= neighZ[:, 1]
            neighZ[:, -1] &= neighZ[:, -2]
            neighZ[:, 1:-1] &= (neighZ[:, :-2] | neighZ[:, 2:])
            crop = neighR & neighZ

    return crop, key, thresh_in


# #############################################################################
# #############################################################################
#                           check
# #############################################################################


def _crop_check(
    coll=None,
    key=None,
    crop=None,
    thresh_in=None,
    remove_isolated=None,
):

    # key
    lkm = list(coll.dobj[coll._which_mesh].keys())
    key = ds._generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=lkm,
    )
    meshtype = coll.dobj[coll._which_mesh][key]['type']

    if meshtype != 'rect':
        raise NotImplementedError()

    # shape
    shape = coll.dobj[coll._which_mesh][key]['shape-c']

    # crop
    c0 = (
        isinstance(crop, np.ndarray)
        and crop.ndim == 2
        and np.all(np.isfinite(crop))
        and (
            (
                crop.shape[0] == 2
                and np.allclose(crop[:, 0], crop[:, -1])
                and (
                    np.issubdtype(crop.dtype, np.integer)
                    or np.issubdtype(crop.dtype, np.floating)
                )
            )
            or (
                crop.shape == shape
                and crop.dtype == np.bool_
            )
        )
    )
    if not c0:
        msg = (
            "Arg crop must be either:\n"
            f"\t- array of bool: mask of shape {shape}\n"
            f"\t- array of floats: (2, npts) closed (R, Z) polygon\n"
            f"Provided:\n{crop}"
        )
        raise Exception(msg)

    cropbool = crop.dtype == np.bool_

    # thresh_in and maxth
    if thresh_in is None:
        thresh_in = 3
    maxth = 5 if coll.dobj[coll._which_mesh][key]['type'] == 'rect' else 4

    c0 = isinstance(thresh_in, (int, np.integer)) and (1 <= thresh_in <= maxth)
    if not c0:
        msg = (
            f"Arg thresh_in must be a int in {1} <= thresh_in <= {maxth}\n"
            f"Provided: {thresh_in}"
        )
        raise Exception(msg)

    # remove_isolated
    remove_isolated = ds._generic_check._check_var(
        remove_isolated, 'remove_isolated',
        default=True,
        types=bool,
    )

    return key, cropbool, thresh_in, remove_isolated


# #############################################################################
# #############################################################################
#                           crop bsplines
# #############################################################################


def add_cropbs_from_crop(coll=None, keybs=None, keym=None):

    # ----------------
    # get

    kcropbs = False
    if coll.dobj[coll._which_mesh][keym]['crop'] is not False:
        kcropm = coll.dobj[coll._which_mesh][keym]['crop']
        cropbs = _get_cropbs_from_crop(
            coll=coll,
            crop=coll.ddata[kcropm]['data'],
            keybs=keybs,
        )
        kcropbs = f'{keybs}_crop'
        kcroppedbs = f'{keybs}_nbs_crop'

    # ----------------
    # optional crop

    if kcropbs is not False:

        # add cropped flat reference
        coll.add_ref(
            key=kcroppedbs,
            size=int(cropbs.sum()),
        )

        coll.add_data(
            key=kcropbs,
            data=cropbs,
            ref=coll._dobj[coll._which_bsplines][keybs]['ref'],
            dim='bool',
            quant='bool',
        )
        coll._dobj[coll._which_bsplines][keybs]['crop'] = kcropbs


# #############################################################################
# #############################################################################
#                           Mesh2DRect - crop
# #############################################################################


def _get_cropbs_from_crop(coll=None, crop=None, keybs=None):

    if isinstance(crop, str) and crop in coll.ddata.keys():
        crop = coll.ddata[crop]['data']

    shref = coll.dobj[coll._which_mesh][coll.dobj['bsplines'][keybs]['mesh']]['shape-c']
    if crop.shape != shref:
        msg = "Arg crop seems to have the wrong shape!"
        raise Exception(msg)

    keym = coll.dobj['bsplines'][keybs][coll._which_mesh]
    kRk, kZk = coll.dobj['mesh'][keym]['knots']
    kRc, kZc = coll.dobj['mesh'][keym]['cents']

    cents_per_bs_R, cents_per_bs_Z = _select._mesh2DRect_bsplines_knotscents(
        returnas='ind',
        return_knots=False,
        return_cents=True,
        ind=None,
        deg=coll.dobj['bsplines'][keybs]['deg'],
        Rknots=coll.ddata[kRk]['data'],
        Zknots=coll.ddata[kZk]['data'],
        Rcents=coll.ddata[kRc]['data'],
        Zcents=coll.ddata[kZc]['data'],
    )

    shapebs = coll.dobj['bsplines'][keybs]['shape']
    cropbs = np.array([
        [
            np.all(crop[cents_per_bs_R[:, ii], cents_per_bs_Z[:, jj]])
            for jj in range(shapebs[1])
        ]
        for ii in range(shapebs[0])
    ], dtype=bool)

    return cropbs
